# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from collections.abc import Sequence
from typing import Any
from typing import Optional

import datarobot as dr
from airflow.exceptions import AirflowFailException
from airflow.utils.context import Context
from datarobot.utils.waiters import wait_for_async_resolution

from datarobot_provider.operators.base_datarobot_operator import BaseDatarobotOperator

DEFAULT_MAX_WAIT_SEC = 600


class CreateExternalModelPackageOperator(BaseDatarobotOperator):
    """
    Create an external model package in DataRobot MLOps from JSON configuration

    :param model_package_json: A JSON object of external model parameters.
    :type model_package_json: dict
    :return: A model package ID of newly created ModelPackage.
    :rtype: str
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = [
        "model_package_json",
    ]

    def __init__(
        self,
        *,
        model_package_json: Optional[dict[str, Any]] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.model_package_json = model_package_json

    # Utility method to support old Model Registry API
    def create_model_package_from_json(self) -> str:
        response = dr.client.get_client().post(
            "modelPackages/fromJSON/", data=self.model_package_json
        )
        if response.status_code == 201:
            response_json = response.json()
            model_package_id = response_json["id"]
            return model_package_id
        else:
            e_msg = "Server unexpectedly returned status code {}"
            raise AirflowFailException(e_msg.format(response.status_code))

    def execute(self, context: Context) -> str:
        if self.model_package_json is None:
            # If model_package_json not provided, trying to get it from DAG params:
            self.model_package_json = context["params"].get("model_package_json")

        if self.model_package_json is None:
            raise ValueError("model_package_json is required.")

        model_package_id = self.create_model_package_from_json()

        self.log.info(f"External Model Package Created, model_package_id={model_package_id}")

        return model_package_id


class DeployModelPackageOperator(BaseDatarobotOperator):
    """
    Create a deployment from a DataRobot model package.
    :deployment_name: A human readable label of the deployment.
    :deployment_name: str
    :model_package_id: The ID of the DataRobot model package to deploy.
    :model_package_id: str
    :default_prediction_server_id: an identifier of a prediction server to be used as the default prediction server
        When working with prediction environments, default prediction server Id should not be provided
    :default_prediction_server_id: str, optional
    :prediction_environment_id:  An identifier of a prediction environment to be used for model deployment.
    :prediction_environment_id: str, optional
    :description: A human readable description of the deployment.
    :description: str, optional
    :importance: Deployment importance level.
    :importance: str, optional
    :user_provided_id: A user-provided unique ID associated with a deployment definition in a remote git repository.
    :user_provided_id: str, optional
    :additional_metadata: A Key/Value pair dict, with additional metadata
    :additional_metadata: dict[str, str], optional
    :max_wait: The amount of seconds to wait for successful resolution of a deployment creation job.
        Deployment supports making predictions only after a deployment creating job has successfully finished.
    :max_wait: int, optional
    :return: The created deployment ID
    :rtype: str
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = [
        "deployment_name",
        "model_package_id",
        "default_prediction_server_id",
        "prediction_environment_id",
        "description",
        "importance",
        "user_provided_id",
        "additional_metadata",
    ]

    def __init__(
        self,
        *,
        deployment_name: str,
        model_package_id: str,
        default_prediction_server_id: Optional[str] = None,
        prediction_environment_id: Optional[str] = None,
        description: Optional[str] = None,
        importance: Optional[str] = None,
        user_provided_id: Optional[str] = None,
        additional_metadata: Optional[dict[str, str]] = None,
        max_wait_sec: int = DEFAULT_MAX_WAIT_SEC,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.deployment_name = deployment_name
        self.model_package_id = model_package_id
        self.default_prediction_server_id = default_prediction_server_id
        self.prediction_environment_id = prediction_environment_id
        self.description = description
        self.importance = importance
        self.user_provided_id = user_provided_id
        self.additional_metadata = additional_metadata
        self.max_wait_sec = max_wait_sec

    # Utility method to support creating Deployment from Model Package
    # Will be refactored after release 3.3 of DataRobot public python client
    @classmethod
    def _create_from_model_package(
        cls,
        model_package_id: str,
        deployment_name: str,
        description: Optional[str] = None,
        default_prediction_server_id: Optional[str] = None,
        prediction_environment_id: Optional[str] = None,
        importance: Optional[str] = None,
        user_provided_id: Optional[str] = None,
        additional_metadata: Optional[dict[str, str]] = None,
        max_wait_sec: int = DEFAULT_MAX_WAIT_SEC,
    ) -> str:
        deployment_payload: dict[str, Any] = {
            "model_package_id": model_package_id,
            "label": deployment_name,
            "description": description,
        }
        if default_prediction_server_id and prediction_environment_id:
            raise ValueError(
                "When working with prediction environments, default prediction server Id should not be provided"
            )
        elif default_prediction_server_id and prediction_environment_id is None:
            deployment_payload["default_prediction_server_id"] = default_prediction_server_id
        elif prediction_environment_id and default_prediction_server_id is None:
            deployment_payload["prediction_environment_id"] = prediction_environment_id

        if importance:
            deployment_payload["importance"] = importance
        if user_provided_id:
            deployment_payload["user_provided_id"] = user_provided_id
        if additional_metadata:
            deployment_payload["additional_metadata"] = additional_metadata
        dr_client = dr.client.get_client()

        response = dr_client.post("deployments/fromModelPackage/", data=deployment_payload)

        if response.status_code == 202 and "Location" in response.headers:
            wait_for_async_resolution(dr_client, response.headers["Location"], max_wait_sec)
            deployment_id = response.json()["id"]
            return deployment_id
        else:
            e_msg = "Server unexpectedly returned status code {}"
            raise AirflowFailException(e_msg.format(response.status_code))

    def validate(self) -> None:
        if self.deployment_name is None:
            raise ValueError("deployment_name is required.")

        if self.model_package_id is None:
            raise ValueError("model_package_id is required.")

    def execute(self, context: Context) -> str:
        if self.additional_metadata is None:
            # If additional_metadata not provided, trying to get it from DAG params:
            self.additional_metadata = context["params"].get("additional_metadata")

        deployment_id = self._create_from_model_package(
            model_package_id=self.model_package_id,
            deployment_name=self.deployment_name,
            description=self.description,
            default_prediction_server_id=self.default_prediction_server_id,
            prediction_environment_id=self.prediction_environment_id,
            importance=self.importance,
            user_provided_id=self.user_provided_id,
            additional_metadata=self.additional_metadata,
            max_wait_sec=self.max_wait_sec,
        )

        self.log.info(f"Deployment Created, deployment_id={deployment_id}")

        return deployment_id



# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from collections.abc import Sequence
from typing import TYPE_CHECKING
from typing import Any
from typing import Optional

import datarobot as dr
from airflow.utils.context import Context

from datarobot_provider.operators.base_datarobot_operator import BaseDatarobotOperator

if TYPE_CHECKING:
    from datarobot.models.deployment.deployment import BiasAndFairnessSettings


class GetBiasAndFairnessSettingsOperator(BaseDatarobotOperator):
    """
    Get Bias And Fairness settings for deployment.

    :param deployment_id: DataRobot deployment ID
    :type deployment_id: str
    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = ["deployment_id"]

    def __init__(
        self,
        *,
        deployment_id: str,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.deployment_id = deployment_id

    def execute(self, context: Context) -> Optional["BiasAndFairnessSettings"]:
        self.log.info(f"Get Deployment for deployment_id={self.deployment_id}")
        deployment = dr.Deployment.get(deployment_id=self.deployment_id)

        bias_and_fairness_settings = deployment.get_bias_and_fairness_settings()

        return bias_and_fairness_settings


class UpdateBiasAndFairnessSettingsOperator(BaseDatarobotOperator):
    """
    Update Bias And Fairness settings for deployment.

    :param deployment_id: DataRobot deployment ID
    :type deployment_id: str
    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = ["deployment_id"]

    def __init__(
        self,
        *,
        deployment_id: str,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.deployment_id = deployment_id

    def execute(self, context: Context) -> None:
        self.log.info(f"Getting Deployment for deployment_id={self.deployment_id}")
        deployment = dr.Deployment.get(deployment_id=self.deployment_id)

        current_bias_and_fairness_settings = deployment.get_bias_and_fairness_settings() or {
            "protected_features": [],
            "fairness_metric_set": "",
            "fairness_threshold": 0.0,
            "preferable_target_value": False,
        }

        protected_features = context["params"].get(
            "protected_features", current_bias_and_fairness_settings["protected_features"]
        )

        fairness_metric_set = context["params"].get(
            "fairness_metric_set",
            # Backward compatibility with parameter name fairness_metrics_set
            context["params"].get(
                "fairness_metrics_set", current_bias_and_fairness_settings["fairness_metric_set"]
            ),
        )

        fairness_threshold = context["params"].get(
            "fairness_threshold", current_bias_and_fairness_settings["fairness_threshold"]
        )

        preferable_target_value = context["params"].get(
            "preferable_target_value", current_bias_and_fairness_settings["preferable_target_value"]
        )

        if (
            (protected_features != current_bias_and_fairness_settings["protected_features"])
            or (fairness_metric_set != current_bias_and_fairness_settings["fairness_metric_set"])
            or (fairness_threshold != current_bias_and_fairness_settings["fairness_threshold"])
            or (
                preferable_target_value
                != current_bias_and_fairness_settings["preferable_target_value"]
            )
        ):
            self.log.debug(
                f"Trying to update bias and fairness settings for deployment_id={self.deployment_id}"
            )
            deployment.update_bias_and_fairness_settings(
                protected_features=protected_features,
                fairness_metric_set=fairness_metric_set,
                fairness_threshold=fairness_threshold,
                preferable_target_value=preferable_target_value,
            )
            self.log.info(
                f"Deployment bias and fairness settings updated for deployment_id={self.deployment_id}"
            )
        else:
            self.log.info(
                f"No need to update bias and fairness settings for deployment_id={self.deployment_id}"
            )



# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from typing import Any

import datarobot as dr
from airflow.sensors.base import BaseSensorOperator
from airflow.utils.context import Context

from datarobot_provider.constants import DATAROBOT_CONN_ID
from datarobot_provider.hooks.datarobot import DataRobotHook


class WaitForPredictionsSensor(BaseSensorOperator):
    """
    Checks whether a predictions job is complete.

    Args:
        project_id (str): DataRobot project ID.
        predict_job_id (str): Predict job ID.

    Returns:
        bool: False if not yet completed, True when complete.
    """

    template_fields = ["project_id", "predict_job_id"]

    def __init__(
        self,
        *,
        project_id: str,
        predict_job_id: str,
        datarobot_conn_id: str = DATAROBOT_CONN_ID,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.predict_job_id = predict_job_id

        self.datarobot_conn_id = datarobot_conn_id
        self.hook = DataRobotHook(datarobot_conn_id)

    def poke(self, context: Context) -> bool:
        # Initialize DataRobot client
        DataRobotHook(datarobot_conn_id=self.datarobot_conn_id).run()

        try:
            self.log.info("Checking if PredictJob is finished.")
            dr.PredictJob.get(project_id=self.project_id, predict_job_id=self.predict_job_id)
        except dr.errors.PendingJobFinished:
            return True
        return False



# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.

import inspect
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional

from airflow.utils.context import Context

from datarobot_provider.operators.base_datarobot_operator import BaseDatarobotOperator


class CustomFunctionOperator(BaseDatarobotOperator):
    """
    Executes a custom function and returns its result.

    Args:
        custom_function (int): Custom function to execute
        params (dict): Keyword arguments to pass to the custom function

    Returns:
        any: Result of the custom function execution

    Raises:
        AirflowFailException: If custom_function is not provided
    """

    template_fields: List[str] = ["func_params"]

    def __init__(
        self,
        *,
        custom_func: Callable,
        func_params: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.custom_func = custom_func
        self.func_params = func_params or {}

    def execute(self, context: Context) -> Any:
        self.log.info(f"Function execution parameters: {self.func_params}")
        if "log" in inspect.signature(self.custom_func).parameters:
            self.func_params["log"] = self.log
        result = self.custom_func(**self.func_params)
        self.log.info(f"Function execution result: {result}")
        return result



# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import logging
from datetime import datetime

from airflow.decorators import dag
from airflow.decorators import task
from datarobot import SCORING_TYPE

from datarobot_provider.operators.model_training import RetrainModelOperator
from datarobot_provider.sensors.model_training import ModelTrainingJobSensor


@dag(
    schedule=None,
    start_date=datetime(2023, 1, 1),
    tags=["example", "retrain"],
    # Default json config example:
    params={
        "sample_pct": 60,  # training dataset sample size to retrain the model.
        "scoring_type": SCORING_TYPE.cross_validation,  # validation type to retrain the model.
    },
)
def retrain_model(
    project_id=None,
    model_id=None,
):
    if not project_id:
        raise ValueError("Invalid or missing `project_id` value")
    if not model_id:
        raise ValueError("Invalid or missing `model_id` value")

    request_model_retraining_op = RetrainModelOperator(
        task_id="request_model_retraining",
        project_id=project_id,
        model_id=model_id,
    )

    model_retraining_sensor = ModelTrainingJobSensor(
        task_id="model_retraining_complete",
        project_id=project_id,
        job_id=request_model_retraining_op.output,
        poke_interval=5,
        timeout=3600,
    )

    @task(task_id="example_custom_python_code")
    def using_custom_python_code(retrained_model_id):
        """Example of using custom python code:"""
        logging.info(msg=f"New retrained model id is: {retrained_model_id}")

    example_custom_python_code = using_custom_python_code(
        retrained_model_id=model_retraining_sensor.output
    )

    request_model_retraining_op >> model_retraining_sensor >> example_custom_python_code


retrain_model_dag = retrain_model()

if __name__ == "__main__":
    retrain_model_dag.test()



# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.

from typing import Any
from typing import Dict
from typing import List

WRANGLER_EXAMPLE_RECIPE: List[Dict[str, Any]] = [
    {
        "directive": "rename-columns",
        "arguments": {
            "columnMappings": [
                {"originalName": "admission_type_id", "newName": "AdmissionType"},
                {
                    "originalName": "discharge_disposition_id",
                    "newName": "DischargeDisposition",
                },
                {"originalName": "admission_source_id", "newName": "AdmissionSource"},
            ]
        },
    },
    {
        "directive": "replace",
        "arguments": {
            "origin": "AdmissionType",
            "searchFor": "",
            "replacement": "Not Available",
            "matchMode": "exact",
        },
    },
    {
        "directive": "replace",
        "arguments": {
            "searchFor": r"\\[(\\d+).*",
            "replacement": r"$1",
            "origin": "age",
            "matchMode": "regex",
        },
    },
    {
        "directive": "compute-new",
        "arguments": {
            "expression": "CAST(`age` AS Integer)",
            "newFeatureName": "int-age",
        },
    },
    {
        "directive": "drop-columns",
        "arguments": {
            "columns": [
                "citoglipton",
                "acetohexamide",
                "miglitol",
                "troglitazone",
                "examide",
                "age",
            ]
        },
    },
]



# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from typing import Any

from airflow.exceptions import AirflowException
from airflow.sensors.base import BaseSensorOperator
from airflow.utils.context import Context
from datarobot._experimental.models.notebooks.enums import ScheduledRunStatus
from datarobot._experimental.models.notebooks.notebook import Notebook
from datarobot._experimental.models.notebooks.scheduled_job import NotebookScheduledJob

from datarobot_provider.constants import DATAROBOT_CONN_ID
from datarobot_provider.hooks.datarobot import DataRobotHook


class NotebookRunCompleteSensor(BaseSensorOperator):
    """
    Checks if the notebook run is complete.

    :param notebook_id: Notebook ID
    :type notebook_id: str
    :param job_id: Job ID
    :type job_id: str
    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    :return: False if not yet completed
    :rtype: bool
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields = [
        "notebook_id",
        "job_id",
    ]

    def __init__(
        self,
        *,
        notebook_id: str = "{{ params.notebook_id }}",
        job_id: str,
        datarobot_conn_id: str = DATAROBOT_CONN_ID,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.notebook_id = notebook_id
        self.job_id = job_id
        self.datarobot_conn_id = datarobot_conn_id

        self.hook = DataRobotHook(datarobot_conn_id)

    def _construct_results_url(self, domain: str, use_case_id: str, revision_id: str) -> str:
        return (
            f"{domain}/usecases/{use_case_id}/overview/notebooks/run-history"
            f"?notebookId={self.notebook_id}"
            f"&revisionId={revision_id}"
        )

    def poke(self, context: Context) -> bool:
        # Initialize DataRobot client
        dr_client = DataRobotHook(datarobot_conn_id=self.datarobot_conn_id).run()

        self.log.info("Checking if notebook execution is complete.")
        notebook = Notebook.get(self.notebook_id)

        # This should not happen but to keep mypy happy as well as remain very defensive we'll check
        if not notebook.use_case_id:
            raise AirflowException("Notebook should have use_case_id")

        notebook_scheduled_job = NotebookScheduledJob.get(
            use_case_id=notebook.use_case_id, scheduled_job_id=self.job_id
        )
        manual_run = notebook_scheduled_job.get_most_recent_run()

        if manual_run and manual_run.status in ScheduledRunStatus.terminal_statuses():
            self.log.info(f"Notebook job no longer running - status: {manual_run.status}")
            if manual_run.revision and manual_run.revision.id:
                results_url = self._construct_results_url(
                    dr_client.domain, notebook.use_case_id, manual_run.revision.id
                )
                self.log.info(f"To view results please visit: {results_url}")
            return True

        status_message = f" - status: {manual_run.status}" if manual_run else ""
        self.log.info(f"Notebook job still running{status_message}")
        return False



# Copyright 2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from datetime import datetime

from airflow.decorators import dag

from datarobot_provider.operators.datarobot import CreateProjectOperator
from datarobot_provider.operators.datarobot import TrainModelsOperator
from datarobot_provider.operators.deployment import DeployRecommendedModelOperator
from datarobot_provider.operators.deployment import GetFeatureDriftOperator
from datarobot_provider.operators.deployment import GetTargetDriftOperator
from datarobot_provider.operators.deployment import ScorePredictionsOperator
from datarobot_provider.sensors.datarobot import AutopilotCompleteSensor
from datarobot_provider.sensors.datarobot import ScoringCompleteSensor


@dag(
    schedule=None,
    start_date=datetime(2022, 1, 1),
    tags=["example"],
)
def datarobot_pipeline():
    create_project_op = CreateProjectOperator(task_id="create_project")

    train_models_op = TrainModelsOperator(
        task_id="train_models",
        project_id=create_project_op.output,
    )

    autopilot_complete_sensor = AutopilotCompleteSensor(
        task_id="check_autopilot_complete",
        project_id=create_project_op.output,
    )

    deploy_model_op = DeployRecommendedModelOperator(
        task_id="deploy_recommended_model",
        project_id=create_project_op.output,
    )

    score_predictions_op = ScorePredictionsOperator(
        task_id="score_predictions",
        deployment_id=deploy_model_op.output,
    )

    scoring_complete_sensor = ScoringCompleteSensor(
        task_id="check_scoring_complete",
        job_id=score_predictions_op.output,
    )

    target_drift_op = GetTargetDriftOperator(
        task_id="target_drift",
        deployment_id=deploy_model_op.output,
    )

    feature_drift_op = GetFeatureDriftOperator(
        task_id="feature_drift",
        deployment_id=deploy_model_op.output,
    )

    (
        create_project_op
        >> train_models_op
        >> autopilot_complete_sensor
        >> deploy_model_op
        >> score_predictions_op
        >> scoring_complete_sensor
    )

    scoring_complete_sensor >> target_drift_op
    scoring_complete_sensor >> feature_drift_op


datarobot_pipeline_dag = datarobot_pipeline()



#
# Copyright 2021-2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import re
from datetime import datetime

from setuptools import find_packages
from setuptools import setup

from setup import DESCRIPTION_TEMPLATE
from setup import common_setup_kwargs
from setup import version

version += datetime.today().strftime(".%Y.%m.%d")

classifiers = [
    "Development Status :: 4 - Beta",
    "Intended Audience :: Developers",
    "Topic :: Scientific/Engineering :: Artificial Intelligence",
    "License :: Other/Proprietary License",
    "Operating System :: OS Independent",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3 :: Only",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3.10",
    "Programming Language :: Python :: 3.11",
    "Programming Language :: Python :: 3.12",
]

# RELEASE SETUP
package_name = "airflow_provider_datarobot_early_access"
pip_package_name = "airflow-provider-datarobot-early-access"

description = DESCRIPTION_TEMPLATE.format(
    package_name=package_name,
    pypi_url_target="https://pypi.python.org/pypi/airflow-provider-datarobot-early-access/",
    extra_desc=(
        'This package is the "early access" version of the client. **Do NOT use this package'
        " in production--you will expose yourself to risk of breaking changes and bugs.** For"
        " the most stable version, see https://pypi.org/project/airflow-provider-datarobot/."
    ),
    pip_package_name=pip_package_name,
)

packages = find_packages(exclude=["docs*", "tests*"])


def _replace_string_in_file(filepath, pattern, replacement):
    with open(filepath, "r") as file:
        file_content = file.read()
    modified_content = re.sub(pattern, replacement, file_content)
    with open(filepath, "w") as file:
        file.write(modified_content)


# Airflow relies on entry point for provider discovery (as seen defined in common_setup_kwargs)
dunder_init = "datarobot_provider/__init__.py"
# Replace package name
mainline = r'"package-name": "airflow-provider-datarobot"'
early_access = '"package-name": "airflow-provider-datarobot-early-access"'
_replace_string_in_file(dunder_init, mainline, early_access)

common_setup_kwargs.update(
    name=package_name,
    version=version,
    packages=packages,
    long_description=description,
    classifiers=classifiers,
    install_requires=["apache-airflow>=2.6.0", "datarobot-early-access"],
)

setup(**common_setup_kwargs)

# Reset changes we made to dunder init file
_replace_string_in_file(dunder_init, early_access, mainline)



# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import logging

logger = logging.getLogger(__package__)

experimental_warning = (
    "You have imported from the _experimental directory.\n"
    "This directory is used for unreleased DataRobot features.\n"
    "Code from this directory may be untested and may change without warning."
)

logger.warning(experimental_warning)



# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from datetime import datetime

from airflow.decorators import dag
from datarobot import AUTOPILOT_MODE

from datarobot_provider.operators.autopilot import StartAutopilotOperator
from datarobot_provider.operators.data_registry import UploadDatasetOperator
from datarobot_provider.operators.datarobot import CreateProjectOperator
from datarobot_provider.sensors.datarobot import AutopilotCompleteSensor


@dag(
    schedule=None,
    start_date=datetime(2022, 1, 1),
    tags=["example", "timeseries"],
    params={
        "dataset_file_path": "/dataset.csv",
        "project_name": "test airflow project timeseries",
        "autopilot_settings": {
            "target": "y",
            "mode": AUTOPILOT_MODE.QUICK,
        },
        "datetime_partitioning_settings": {
            "use_time_series": True,
            "datetime_partition_column": "datetime",
            "multiseries_id_columns": ["location"],
        },
        "unsupervised_mode": False,
        "use_feature_discovery": False,
    },
)
def datarobot_timeseries_pipeline():
    dataset_uploading_op = UploadDatasetOperator(
        task_id="dataset_uploading",
    )

    create_project_op = CreateProjectOperator(
        task_id="create_project",
        dataset_id=dataset_uploading_op.output,
    )

    train_models_op = StartAutopilotOperator(
        task_id="train_timeseries_models",
        project_id=create_project_op.output,
    )

    autopilot_complete_sensor = AutopilotCompleteSensor(
        task_id="check_autopilot_complete",
        project_id=create_project_op.output,
    )

    dataset_uploading_op >> create_project_op >> train_models_op >> autopilot_complete_sensor


datarobot_timeseries_pipeline_dag = datarobot_timeseries_pipeline()

if __name__ == "__main__":
    datarobot_timeseries_pipeline_dag.test()



# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from datetime import datetime

from airflow.decorators import dag
from airflow.decorators import task

from datarobot_provider.operators.monitoring import GetServiceStatsOperator


@dag(
    schedule=None,
    start_date=datetime(2023, 1, 1),
    tags=["example", "mlops"],
)
def deployment_service_stats():
    service_stats_op = GetServiceStatsOperator(
        task_id="get_service_stats",
        # you can pass deployment_id from previous operator here:
        deployment_id="put your deployment_id here",
    )

    @task(task_id="example_processing_python")
    def service_stat_processing(model_service_stat):
        """Example of custom logic based on service stats from the deployment."""

        # Put your service stat processing logic here:
        current_model_id = model_service_stat["model_id"]
        total_predictions = model_service_stat["metrics"]["totalPredictions"]
        print(f"model_id:{current_model_id}, total_predictions:{total_predictions}")

        return total_predictions

    example_service_stat_processing = service_stat_processing(
        model_service_stat=service_stats_op.output
    )

    service_stats_op >> example_service_stat_processing


deployment_service_stats_dag = deployment_service_stats()

if __name__ == "__main__":
    deployment_service_stats_dag.test()



# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.

from datetime import datetime

from airflow.decorators import dag
from datarobot import TARGET_TYPE
from datarobot.enums import DEPLOYMENT_IMPORTANCE

from datarobot_provider.operators.model_package import CreateExternalModelPackageOperator
from datarobot_provider.operators.model_package import DeployModelPackageOperator
from datarobot_provider.operators.monitoring import UpdateMonitoringSettingsOperator

"""
Example of JSON configuration for a regression model:

.. code-block:: python

    {
         "name": "Lending club regression",
         "modelDescription": {
                 "description": "Regression on lending club dataset"
             }
         "target": {
             "type": "Regression",
             "name": "loan_amnt"
         }
    }


Example JSON for a binary classification model:

.. code-block:: python

    {
        "name": "Surgical Model",
        "modelDescription": {
            "description": "Binary classification on surgical dataset",
            "location": "/tmp/myModel"
            },
            "target": {
                "type": "Binary",
                "name": "complication",
                "classNames": ["Yes","No"],  # minority/positive class should be listed first
                "predictionThreshold": 0.5
            }
        }
    }

Example JSON for a multiclass classification model:

.. code-block:: python

    {
        "name": "Iris classifier",
        "modelDescription": {
        "description": "Classification on iris dataset",
        "location": "/tmp/myModel"
    },
        "target": {
            "type": "Multiclass",
            "name": "Species",
            "classNames": [
                "Iris-versicolor",
                "Iris-virginica",
                "Iris-setosa"
            ]
        }
    }
"""


@dag(
    schedule=None,
    start_date=datetime(2023, 1, 1),
    tags=["example", "mlops", "external model"],
    params={
        "model_package_json": {
            "name": "Demo Regression Model",
            "modelDescription": {"description": "Regression on demo dataset"},
            "target": {"type": TARGET_TYPE.REGRESSION, "name": "Grade 2014"},
        },
        "target_drift_enabled": True,
        "feature_drift_enabled": True,
        "association_id_column": ["id"],
        "required_association_id": False,
        "predictions_data_collection_enabled": False,
    },
)
def create_external_deployment_pipeline(prediction_environment_id=None):
    if not prediction_environment_id:
        raise ValueError("Invalid or missing `prediction_environment_id` value")

    create_model_package_op = CreateExternalModelPackageOperator(
        task_id="create_model_package",
    )

    deploy_model_package_op = DeployModelPackageOperator(
        task_id="deployment_from_model_package",
        deployment_name="demo_airflow_deployment",
        description="demo_airflow_deployment",
        model_package_id=create_model_package_op.output,
        prediction_environment_id=prediction_environment_id,
        importance=DEPLOYMENT_IMPORTANCE.LOW,
    )

    update_monitoring_settings_op = UpdateMonitoringSettingsOperator(
        task_id="update_monitoring_settings",
        deployment_id=deploy_model_package_op.output,
    )

    create_model_package_op >> deploy_model_package_op >> update_monitoring_settings_op


create_external_deployment_pipeline_dag = create_external_deployment_pipeline()

if __name__ == "__main__":
    create_external_deployment_pipeline_dag.test()



# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
"""
Example of DAG to upload local file to DataRobot AI Catalog as a dataset version,
then using this dataset version as a training data to create a DataRobot Project.

Config example for this dag:
{
    "training_dataset_id": "existing training dataset id"
    "dataset_file_path": "/path/to/local/file",
    "project_name": "test project",
    "unsupervised_mode": False,
    "use_feature_discovery": False
}
"""

from datetime import datetime

from airflow.decorators import dag

from datarobot_provider.operators.data_registry import UpdateDatasetFromFileOperator
from datarobot_provider.operators.datarobot import CreateProjectOperator


@dag(
    schedule=None,
    start_date=datetime(2022, 1, 1),
    tags=["example"],
    params={
        "training_dataset_id": "644...590",
        "dataset_file_path": "/path/to/local/file",
        "project_name": "test project updated",
        "unsupervised_mode": False,
        "use_feature_discovery": False,
    },
)
def create_project_from_dataset_version():
    dataset_new_version_op = UpdateDatasetFromFileOperator(
        task_id="dataset_new_version",
    )

    create_project_op = CreateProjectOperator(
        task_id="create_project", dataset_version_id=dataset_new_version_op.output
    )

    dataset_new_version_op >> create_project_op


create_project_from_dataset_version_dag = create_project_from_dataset_version()



# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from datetime import datetime

import datarobot as dr
from airflow.decorators import dag
from airflow.decorators import task

from datarobot_provider.hooks.datarobot import DataRobotHook
from datarobot_provider.operators.model_insights import ComputeFeatureImpactOperator
from datarobot_provider.operators.model_predictions import AddExternalDatasetOperator
from datarobot_provider.operators.model_predictions import RequestModelPredictionsOperator
from datarobot_provider.operators.prediction_explanations import (
    ComputePredictionExplanationsOperator,
)
from datarobot_provider.operators.prediction_explanations import (
    PredictionExplanationsInitializationOperator,
)
from datarobot_provider.sensors.model_insights import DataRobotJobSensor

"""
Example DAG to create prediction explanations for the specified dataset.
In order to create Prediction Explanations for a particular model and dataset, you must use the next steps:
 - Compute feature impact for the model using ComputeFeatureImpactOperator
 - Initialize prediction explanations for the model using PredictionExplanationsInitializationOperator
 - Compute predictions for the model and dataset using RequestModelPredictionsOperator
 - Compute predictions explanations for the model and dataset using ComputePredictionExplanationsOperator
"""


@dag(
    schedule=None,
    start_date=datetime(2023, 1, 1),
    tags=["example", "dataset", "model"],
    # Default json config example:
    params={"threshold_high": 0.9, "threshold_low": 0.1, "max_explanations": 3},
)
def compute_model_prediction_explanations(
    project_id=None,
    model_id=None,
    dataset_id=None,
):
    if not project_id:
        raise ValueError("Invalid or missing `project_id` value")
    if not model_id:
        raise ValueError("Invalid or missing `model_id` value")
    if not dataset_id:
        raise ValueError("Invalid or missing `dataset_id` value")

    add_external_dataset_op = AddExternalDatasetOperator(
        task_id="add_external_dataset",
        project_id=project_id,
        dataset_id=dataset_id,
    )

    compute_feature_impact_op = ComputeFeatureImpactOperator(
        task_id="compute_feature_impact",
        project_id=project_id,
        model_id=model_id,
    )

    feature_impact_complete_sensor = DataRobotJobSensor(
        task_id="feature_impact_complete",
        project_id=project_id,
        job_id=compute_feature_impact_op.output,
        poke_interval=5,
        timeout=3600,
    )

    request_model_predictions_op = RequestModelPredictionsOperator(
        task_id="request_model_predictions",
        project_id=project_id,
        model_id=model_id,
        external_dataset_id=add_external_dataset_op.output,
    )

    model_predictions_sensor = DataRobotJobSensor(
        task_id="model_predictions_complete",
        project_id=project_id,
        job_id=request_model_predictions_op.output,
        poke_interval=5,
        timeout=3600,
    )

    # In order to compute prediction explanations you have to initialize it for a particular model.
    prediction_explanations_initialization_op = PredictionExplanationsInitializationOperator(
        task_id="prediction_explanations_initialization",
        project_id=project_id,
        model_id=model_id,
    )

    prediction_explanations_initialization_sensor = DataRobotJobSensor(
        task_id="prediction_explanations_initialization_complete",
        project_id=project_id,
        job_id=prediction_explanations_initialization_op.output,
        poke_interval=5,
        timeout=3600,
    )

    compute_prediction_explanations_op = ComputePredictionExplanationsOperator(
        task_id="compute_prediction_explanations",
        project_id=project_id,
        model_id=model_id,
        external_dataset_id=add_external_dataset_op.output,
    )

    compute_prediction_explanations_sensor = DataRobotJobSensor(
        task_id="compute_prediction_explanations_complete",
        project_id=project_id,
        job_id=compute_prediction_explanations_op.output,
        poke_interval=5,
        timeout=3600,
    )

    @task(task_id="example_custom_python_code")
    def using_custom_python_code(
        pe_project_id, pe_model_id, predict_job_id, datarobot_conn_id="datarobot_default"
    ):
        """Example of using custom python code:"""

        # Initialize DataRobot client
        DataRobotHook(datarobot_conn_id).run()

        # Should be used only with a very small datasets (not higher than a few megabytes).
        predictions_df = dr.PredictJob.get_predictions(
            project_id=pe_project_id, predict_job_id=predict_job_id
        )
        # Get last prediction explanations computation
        explanations_id = dr.PredictionExplanations.list(pe_project_id, model_id=pe_model_id)[-1].id
        prediction_explanations_df = dr.PredictionExplanations.get(
            pe_project_id, explanations_id
        ).get_all_as_dataframe()

        # Put your logic with the model predictions output and prediction explanations here, for example:
        return predictions_df["positive_probability"].mean(), prediction_explanations_df[
            "explanation_0_feature"
        ].describe().get("top")

    example_custom_python_code = using_custom_python_code(
        pe_project_id=project_id,
        pe_model_id=model_id,
        predict_job_id=request_model_predictions_op.output,
    )

    (
        add_external_dataset_op
        >> compute_feature_impact_op
        >> feature_impact_complete_sensor
        >> prediction_explanations_initialization_op
        >> prediction_explanations_initialization_sensor
        >> request_model_predictions_op
        >> model_predictions_sensor
        >> compute_prediction_explanations_op
        >> compute_prediction_explanations_sensor
        >> example_custom_python_code
    )


compute_model_prediction_explanations_dag = compute_model_prediction_explanations()

if __name__ == "__main__":
    compute_model_prediction_explanations_dag.test()



# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.

from typing import Any
from typing import Optional
from typing import Union

import datarobot as dr
from airflow.exceptions import AirflowException
from airflow.models import BaseOperator
from airflow.utils.context import Context

from datarobot_provider.constants import DATAROBOT_CONN_ID
from datarobot_provider.constants import XCOM_DEFAULT_USE_CASE_ID
from datarobot_provider.hooks.datarobot import DataRobotHook


class BaseDatarobotOperator(BaseOperator):
    ui_color = "#f4a460"

    dr_hook: DataRobotHook

    def pre_execute(self, context: Context) -> None:
        super().pre_execute(context)

        self.dr_hook = DataRobotHook(datarobot_conn_id=self.datarobot_conn_id)
        self.dr_hook.run()

        self.validate()

    def validate(self) -> None:
        """Implement your validation of rendered operator fields here."""

    def __init__(self, *, datarobot_conn_id: str = DATAROBOT_CONN_ID, **kwargs: Any):
        super().__init__(**kwargs)
        self.datarobot_conn_id = datarobot_conn_id
        if kwargs.get("xcom_push") is not None:
            raise AirflowException(
                "'xcom_push' was deprecated, use 'BaseOperator.do_xcom_push' instead"
            )


class BaseUseCaseEntityOperator(BaseDatarobotOperator):
    """Base class for all operators which create a Use Case assets,
    i.e. dataset, recipe, project, deployment, notebook, application, etc.

    It introduces *use_case_id* template parameter and helper methods
    to use a default Use Case if no Use Case was defined in the operator itself or a DAG context.
    """

    @classmethod
    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)
        if "use_case_id" not in cls.template_fields:
            cls.template_fields = (*cls.template_fields, "use_case_id")
            cls.logger().info(
                "use_case_id is implicitly added into %s template_fields list.", cls.__name__
            )

    def __init__(
        self,
        *,
        use_case_id: Optional[str] = "{{ params.use_case_id | default('') }}",
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.use_case_id = use_case_id

    def get_use_case(self, context: Context, required: bool = False) -> Optional[dr.UseCase]:
        """Get a `dr.UseCase` entity based on self.use_case_id or a default Use Case defined at runtime.
        Raises an exception if no Use Case is defined and required=True
        """
        if use_case_id := self.get_use_case_id(context, required=required):
            return dr.UseCase.get(use_case_id)

        return None

    def get_use_case_id(self, context: Context, required: bool = False) -> Optional[str]:
        """Get self.use_case_id or a default Use Case id defined at runtime.
        Raises an exception if no Use Case id is defined and required=True"""
        if use_case_id := (
            self.use_case_id or self.xcom_pull(context, key=XCOM_DEFAULT_USE_CASE_ID)
        ):
            return use_case_id

        if required:
            raise AirflowException(
                f"{self.__class__.__name__} requires a Use Case. Please define one of:\n"
                "*use_case_id* parameter in the operator\n"
                "*use_case_id* DAG context parameter\n"
                "`GetOrCreateUseCaseOperator(..., set_default=True)` as one of the previous DAG tasks"
            )

        return None

    def add_into_use_case(
        self,
        entity: Union[dr.Project, dr.Dataset, dr.models.Recipe, dr.Application],
        *,
        context: Context,
    ) -> None:
        """Add an *entity* into Use Case defined as self.use_case_id or a default DAG Use Case."""
        if use_case := self.get_use_case(context):
            use_case.add(entity)
            self.log.info(
                '%s is added into use case "%s"', entity.__class__.__name__, use_case.name
            )

        else:
            self.log.info("The %s won't be added into any use case.", entity.__class__.__name__)



# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from json import dumps
from typing import TYPE_CHECKING
from typing import Any
from typing import Dict
from typing import Optional
from typing import Sequence

import datarobot as dr
from airflow.exceptions import AirflowFailException
from airflow.utils.context import Context
from datarobot.enums import MODEL_REPLACEMENT_REASON

from datarobot_provider.operators.base_datarobot_operator import BaseDatarobotOperator
from datarobot_provider.operators.datarobot import DATAROBOT_MAX_WAIT
from datarobot_provider.operators.datarobot import DATETIME_FORMAT

if TYPE_CHECKING:
    from datarobot.models.deployment.deployment import ModelDict


def _serialize_drift(drift_obj, date_format=DATETIME_FORMAT):
    drift_dict = drift_obj.__dict__.copy()
    drift_dict["period"] = {
        "start": drift_obj.period["start"].strftime(date_format),
        "end": drift_obj.period["end"].strftime(date_format),
    }
    return drift_dict


class DeployModelMixin:
    def deploy_model(
        self, model_id: str, label: str, description: Optional[str] = None
    ) -> dr.Deployment:
        """Deploys the provided model to production."""
        self.log.info(f"Deploying model_id={model_id} with label={label}")  # type: ignore[attr-defined]
        prediction_server = dr.PredictionServer.list()[0]
        deployment = dr.Deployment.create_from_learning_model(
            model_id, label, description, prediction_server.id
        )
        self.log.info(f"Model deployed: deployment_id={deployment.id}")  # type: ignore[attr-defined]
        self.log.info("Enabling tracking for target drift and feature drift")  # type: ignore[attr-defined]
        deployment.update_drift_tracking_settings(
            target_drift_enabled=True, feature_drift_enabled=True, max_wait=DATAROBOT_MAX_WAIT
        )
        return deployment


class DeployModelOperator(BaseDatarobotOperator, DeployModelMixin):
    """
    Deploys the specified model to production.

    :param model_id: ID of the DataRobot model to deploy
    :type model_id: str
    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    :return: DataRobot deployment ID
    :rtype: str
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = ["model_id"]

    def __init__(
        self,
        *,
        model_id: str,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.model_id = model_id

    def execute(self, context: Context) -> str:
        # Deploy the model
        deployment = self.deploy_model(
            self.model_id,
            context["params"]["deployment_label"],
            context["params"].get("deployment_description"),
        )
        return deployment.id


class DeployRecommendedModelOperator(BaseDatarobotOperator, DeployModelMixin):
    """
    Deploys a recommended model to production.

    :param project_id: DataRobot project ID
    :type project_id: str
    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    :return: DataRobot deployment ID
    :rtype: str
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = ["project_id"]

    def __init__(
        self,
        *,
        project_id: str,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id

    def deploy_recommended_model(
        self, project_id: str, label: str, description: Optional[str] = None
    ) -> dr.Deployment:
        """Deploys the recommended model to production."""
        self.log.info(f"Retrieving recommended model for project_id={project_id}")
        project: dr.Project = dr.Project.get(project_id)
        model = project.recommended_model()
        if model is None:
            raise AirflowFailException(f"No recommended model found for project_id={project_id}")
        return self.deploy_model(model.id, label, description)

    def execute(self, context: Context) -> str:
        # Deploy the recommended model
        deployment = self.deploy_recommended_model(
            self.project_id,
            context["params"]["deployment_label"],
            context["params"].get("deployment_description"),
        )
        return deployment.id


class GetDeploymentModelOperator(BaseDatarobotOperator):
    """
    Gets current model info from a deployment.

    :param deployment_id: DataRobot deployment ID
    :type deployment_id: str
    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    :return: model info from a Deployment
    :rtype: str
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = ["deployment_id"]

    def __init__(
        self,
        *,
        deployment_id: str,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.deployment_id = deployment_id

    def execute(self, context: Context) -> Optional["ModelDict"]:
        self.log.info(f"Getting model_id for deployment_id={self.deployment_id}")
        deployment = dr.Deployment.get(self.deployment_id)
        return deployment.model


class GetDeploymentStatusOperator(BaseDatarobotOperator):
    """
    Get a Deployment status (active/inactive).

    :param deployment_id: DataRobot deployment ID
    :type deployment_id: str
    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    :return: Deployment status (active/inactive)
    :rtype: str
    """

    # Specify the arguments that are allowed to parse with jinja templating

    template_fields: Sequence[str] = ["deployment_id"]

    def __init__(
        self,
        *,
        deployment_id: str,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.deployment_id = deployment_id

    def validate(self) -> None:
        if not self.deployment_id:
            raise ValueError("Invalid or missing `deployment_id` value")

    def execute(self, context: Context) -> Optional[str]:
        self.log.info(f"Getting Deployment for deployment_id={self.deployment_id}")
        deployment = dr.Deployment.get(self.deployment_id)
        return deployment.status


class ReplaceModelOperator(BaseDatarobotOperator):
    """
    Replaces the current model for a deployment.

    :param deployment_id: DataRobot deployment ID
    :type deployment_id: str
    :param new_model_id: The id of the new model to use. If replacing the deployment's model with a
            CustomInferenceModel, a specific CustomModelVersion ID must be used.
    :type new_model_id: str
    :param reason: str
            The reason for the model replacement. Must be one of 'ACCURACY', 'DATA_DRIFT', 'ERRORS',
            'SCHEDULED_REFRESH', 'SCORING_SPEED', or 'OTHER'. This value will be stored in the model
            history to keep track of why a model was replaced
    :type reason: str, optional
    :param max_wait_sec: The maximum time to wait for
            model replacement job to complete before erroring,
            defaults to 600 seconds
    :type max_wait_sec: int, optional
    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = ["deployment_id", "new_registered_model_version_id", "reason"]

    def __init__(
        self,
        *,
        deployment_id: str,
        new_registered_model_version_id: str,
        reason: str = MODEL_REPLACEMENT_REASON.OTHER,
        max_wait_sec: int = 600,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.deployment_id = deployment_id
        self.new_registered_model_version_id = new_registered_model_version_id
        self.reason = reason
        self.max_wait_sec = max_wait_sec

    def validate(self) -> None:
        if self.deployment_id is None:
            raise AirflowFailException("deployment_id must be provided")

        if self.new_registered_model_version_id is None:
            raise AirflowFailException("new_registered_model_version_id must be provided")

    def execute(self, context: Context) -> None:
        self.log.info(f"Getting model_id for deployment_id={self.deployment_id}")
        deployment = dr.Deployment.get(deployment_id=self.deployment_id)
        self.log.info(
            f"Validating replacement model new_registered_model_version_id={self.new_registered_model_version_id} "
            f"for deployment_id={self.deployment_id}"
        )
        check_result, check_message, status_list = deployment.validate_replacement_model(
            new_registered_model_version_id=self.new_registered_model_version_id
        )
        self.log.info(f"Validation result: {check_result}, message: {check_message}")
        self.log.info("Validation result details:\n%s", dumps(status_list, indent=2))
        if check_result == "failing":
            raise AirflowFailException(check_message)
        self.log.info(
            f"Trying to replace a model for deployment_id={self.deployment_id} "
            f"to new_registered_model_version_id={self.new_registered_model_version_id}"
        )
        deployment.perform_model_replace(
            new_registered_model_version_id=self.new_registered_model_version_id,
            reason=self.reason,
            max_wait=self.max_wait_sec,
        )
        self.log.info(
            f"Model for deployment_id={self.deployment_id} replaced "
            f"to new_registered_model_version_id={self.new_registered_model_version_id}"
        )


class ScorePredictionsOperator(BaseDatarobotOperator):
    """
    Creates a batch prediction job in DataRobot, scores the data and saves prediction to the output.
    :param deployment_id: DataRobot deployment ID
    :type deployment_id: str
    :param intake_datastore_id: DataRobot DataStore ID for jdbc source connection
    :type intake_datastore_id: str
    :param intake_credential_id: DataRobot Credentials ID for source connection
    :type intake_credential_id: str
    :param output_datastore_id: DataRobot DataStore ID for jdbc destination connection
    :type output_datastore_id: str
    :param output_credential_id: DataRobot Credentials ID for destination connection
    :type output_credential_id: str
    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    :return: Batch predictions job ID
    :rtype: str
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = [
        "deployment_id",
        "intake_datastore_id",
        "intake_credential_id",
        "output_datastore_id",
        "output_credential_id",
        "score_settings",
    ]

    def __init__(
        self,
        *,
        deployment_id: Optional[str] = None,
        intake_datastore_id: Optional[str] = None,
        intake_credential_id: Optional[str] = None,
        output_datastore_id: Optional[str] = None,
        output_credential_id: Optional[str] = None,
        score_settings: Optional[dict] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.deployment_id = deployment_id
        self.intake_datastore_id = intake_datastore_id
        self.intake_credential_id = intake_credential_id
        self.output_datastore_id = output_datastore_id
        self.output_credential_id = output_credential_id
        self.score_settings = score_settings

    def execute(self, context: Context) -> str:
        if self.score_settings is None:
            self.score_settings = context["params"]["score_settings"]

        # in case of deployment_id was not set from operator argument:
        if self.deployment_id is None:
            self.deployment_id = self.score_settings["deployment_id"]

        if self.intake_credential_id is not None:
            self.score_settings["intake_settings"]["credential_id"] = self.intake_credential_id
        if self.output_credential_id is not None:
            self.score_settings["output_settings"]["credential_id"] = self.output_credential_id

        intake_settings = self.score_settings.get("intake_settings", dict())
        output_settings = self.score_settings.get("output_settings", dict())

        intake_type = intake_settings.get("type")
        output_type = output_settings.get("type")

        # in case of JDBC intake from operator argument:
        if intake_type == "jdbc" and self.intake_datastore_id is not None:
            self.score_settings["intake_settings"]["data_store_id"] = self.intake_datastore_id

        # in case of JDBC output from operator argument:
        if output_type == "jdbc" and self.output_datastore_id is not None:
            self.score_settings["output_settings"]["data_store_id"] = self.output_datastore_id

        # Score data
        self.log.info(
            f"Scoring predictions against deployment_id={self.deployment_id} "
            f"with settings: {self.score_settings}"
        )

        # BatchPredictionJob.score() method in the Python SDK expects a DataRobot dataset instance
        if intake_type == "dataset":
            dataset_id = intake_settings.get("dataset_id")
            if not dataset_id:
                raise ValueError(
                    "Invalid or missing `dataset_id` value for the `dataset` intake type."
                )
            dataset = dr.Dataset.get(dataset_id)
            intake_settings["dataset"] = dataset

            # We no longer need the ID
            del intake_settings["dataset_id"]

        job = dr.BatchPredictionJob.score(self.deployment_id, **self.score_settings)
        self.log.info(f"Batch Prediction submitted, job.id={job.id}")
        return job.id


class ActivateDeploymentOperator(BaseDatarobotOperator):
    """
    Activate or deactivate a Deployment.

    :param deployment_id: DataRobot deployment ID
    :type deployment_id: str
    :param activate: if set to True - activate deployment, if set to False - deactivate deployment
    :type activate: boolean
    :param max_wait_sec: The maximum time in seconds to wait for deployment activation/deactivation to complete
    :type max_wait_sec: int
    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    :return: Deployment status (active/inactive)
    :rtype: str
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = ["deployment_id", "activate"]

    def __init__(
        self,
        *,
        deployment_id: str,
        activate: bool = True,
        max_wait_sec: int = 600,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.deployment_id = deployment_id
        self.activate = activate
        self.max_wait_sec = max_wait_sec

    def validate(self) -> None:
        if self.deployment_id is None:
            raise ValueError("Invalid or missing `deployment_id` value")

    def execute(self, context: Context) -> Optional[str]:
        self.log.info(f"Getting Deployment with deployment_id={self.deployment_id}")
        deployment = dr.Deployment.get(self.deployment_id)
        if self.activate:
            deployment.activate(max_wait=self.max_wait_sec)
        else:
            deployment.deactivate(max_wait=self.max_wait_sec)
        return deployment.status


class GetFeatureDriftOperator(BaseDatarobotOperator):
    """
    Gets feature drift measurements from a deployment.

    :param deployment_id: DataRobot deployment ID
    :type deployment_id: str
    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    :return: Drift stats for a Deployment's features
    :rtype: list[dict]
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = ["deployment_id"]

    def __init__(
        self,
        *,
        deployment_id: str,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.deployment_id = deployment_id

    def execute(self, context: Context) -> list[dict]:
        self.log.info(f"Getting feature drift for deployment_id={self.deployment_id}")
        deployment = dr.Deployment.get(self.deployment_id)
        feature_drift_params = context["params"].get("feature_drift", {})
        drift = deployment.get_feature_drift(**feature_drift_params)
        return [_serialize_drift(feature) for feature in drift]


class GetTargetDriftOperator(BaseDatarobotOperator):
    """
    Gets target drift measurements from a deployment.

    :param deployment_id: DataRobot deployment ID
    :type deployment_id: str
    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    :return: Drift stats for a Deployment's target
    :rtype: dict
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = ["deployment_id"]

    def __init__(
        self,
        *,
        deployment_id: str,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.deployment_id = deployment_id

    def execute(self, context: Context) -> dict[str, Any]:
        self.log.info(f"Getting target drift for deployment_id={self.deployment_id}")
        deployment = dr.Deployment.get(self.deployment_id)
        target_drift_params = context["params"].get("target_drift", {})
        drift = deployment.get_target_drift(**target_drift_params)
        return _serialize_drift(drift)


class DeployRegisteredModelOperator(BaseDatarobotOperator):
    """
    Create a deployment from a registered model version using DataRobot's API.

    This operator creates a deployment for a registered model version by calling
    DataRobot's `Deployment.create_from_registered_model_version()` method. It allows
    optional extra parameters to be passed to the DataRobot client call.

    Args:
        model_package_id (str): The registered model version ID to deploy.
        deployment_label (str): The label or name to assign to the deployment.
        extra_params (dict, optional): A dictionary of additional parameters to pass
            to the DataRobot deployment creation API.
        kwargs (dict): Additional keyword arguments passed to the BaseDatarobotOperator.
    """

    template_fields: Sequence[str] = ["model_package_id", "deployment_label", "extra_params"]

    def __init__(
        self,
        *,
        model_package_id: str,
        deployment_label: str,
        extra_params: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> None:
        self.extra_params = extra_params or {}
        super().__init__(**kwargs)
        self.model_package_id = model_package_id
        self.deployment_label = deployment_label

    def validate(self) -> None:
        if not self.model_package_id:
            raise ValueError("model_package_id must be provided.")
        if not self.deployment_label:
            raise ValueError("label must be provided.")

    def execute(self, context: Context) -> str:
        self.log.info(
            "Creating deployment from registered model version: %s", self.model_package_id
        )

        # Create deployment from registered model.
        deployment = dr.Deployment.create_from_registered_model_version(
            model_package_id=self.model_package_id, label=self.deployment_label, **self.extra_params
        )

        self.log.info("Deployment created with ID: %s", deployment.id)
        return deployment.id



# Copyright 2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from typing import Any
from typing import Union

from airflow.sensors.base import BaseSensorOperator
from airflow.sensors.base import PokeReturnValue
from airflow.utils.context import Context
from datarobot import Job
from datarobot.errors import AsyncProcessUnsuccessfulError

from datarobot_provider.constants import DATAROBOT_CONN_ID
from datarobot_provider.hooks.datarobot import DataRobotHook


class DataRobotJobSensor(BaseSensorOperator):
    """
    Checks whether DataRobot Job is complete.

    Args:
        project_id (str): DataRobot project ID.
        job_id (str): DataRobot Job ID.
        datarobot_conn_id (str, optional): Connection ID, defaults to `datarobot_default`.

    Returns:
        bool: False if not yet completed.
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields = ["project_id", "job_id"]

    def __init__(
        self,
        *,
        project_id: str,
        job_id: str,
        datarobot_conn_id: str = DATAROBOT_CONN_ID,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.job_id = job_id
        self.datarobot_conn_id = datarobot_conn_id

        self.hook = DataRobotHook(datarobot_conn_id)

    def get_job_result(self, context: Context) -> Union[bool, PokeReturnValue]:
        # Default implementation return True if job is completed:
        return True

    def poke(self, context: Context) -> Union[bool, PokeReturnValue]:
        # Initialize DataRobot client
        DataRobotHook(datarobot_conn_id=self.datarobot_conn_id).run()

        self.log.info("Checking if DataRobot Job is complete")

        job = Job.get(project_id=self.project_id, job_id=self.job_id)

        if job.status.lower() in ["error", "abort"]:
            raise AsyncProcessUnsuccessfulError(
                f"The job did not complete successfully. Job Status: {job.status}"
            )
        if job.status.lower() == "completed":
            return self.get_job_result(context)
        return False



# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import os
from collections.abc import Sequence
from typing import Any
from typing import Optional

import datarobot as dr
from airflow.exceptions import AirflowException
from airflow.utils.context import Context

from datarobot_provider.operators.base_datarobot_operator import BaseDatarobotOperator


class DownloadDeploymentScoringCodeOperator(BaseDatarobotOperator):
    """
    Downloads scoring code from a deployment.

    :param deployment_id: DataRobot deployment ID
    :type deployment_id: str
    :param base_path: base path for storing downloaded model artifact
    :type base_path: str
    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    :return: path to the downloaded jar file
    :rtype: String
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = ["deployment_id", "base_path"]

    def __init__(
        self,
        *,
        deployment_id: str,
        base_path: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.deployment_id = deployment_id
        self.base_path = base_path

    def execute(self, context: Context) -> str:
        self.log.info(f"Getting model scoring code for deployment_id={self.deployment_id}")
        deployment = dr.Deployment.get(self.deployment_id)
        # in case if params provided
        self.base_path = context["params"].get("scoring_code_filepath", self.base_path)
        if self.base_path is None or not os.path.exists(self.base_path):
            raise ValueError("Invalid or missing base path value, make sure that directory exists")

        if not deployment.model:
            raise AirflowException(
                f"Deployment {self.deployment_id} does not have a model associated"
            )

        scoring_code_path = os.path.join(
            self.base_path, f"{deployment.id}-{deployment.model['id']}.jar"
        )

        source_code = context["params"].get("source_code", False)
        include_agent = context["params"].get("include_agent", False)
        include_prediction_explanations = context["params"].get(
            "include_prediction_explanations", False
        )
        include_prediction_intervals = context["params"].get("include_prediction_intervals", False)

        self.log.debug(f"Trying to download scoring code for deployment_id={self.deployment_id}")

        deployment.download_scoring_code(
            filepath=scoring_code_path,
            source_code=source_code,
            include_agent=include_agent,
            include_prediction_explanations=include_prediction_explanations,
            include_prediction_intervals=include_prediction_intervals,
        )

        self.log.info(
            f"Scoring code for deployment_id={self.deployment_id} downloaded to the filepath={scoring_code_path}"
        )

        return scoring_code_path


class DownloadModelScoringCodeOperator(BaseDatarobotOperator):
    """
    Downloads scoring code from a model.

    :param project_id: DataRobot project ID
    :type project_id: str
    :param model_id: DataRobot model ID
    :type model_id: str
    :param base_path: base path for storing downloaded model artifact
    :type base_path: str
    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    :return: path to the downloaded jar file
    :rtype: String
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = ["project_id", "model_id"]

    def __init__(
        self,
        *,
        project_id: str,
        model_id: str,
        base_path: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.model_id = model_id
        self.base_path = base_path

    def execute(self, context: Context) -> str:
        # in case if params provided
        self.base_path = context["params"].get("scoring_code_filepath", self.base_path)
        if self.base_path is None or not os.path.exists(self.base_path):
            raise ValueError("Invalid or missing base path value, make sure that directory exists")

        scoring_code_path = os.path.join(self.base_path, f"{self.model_id}.jar")

        source_code = context["params"].get("source_code", False)

        self.log.info(
            f"Trying to download scoring code for project_id={self.project_id} and model_id={self.model_id}"
        )

        model = dr.Model.get(project=self.project_id, model_id=self.model_id)
        model.download_scoring_code(file_name=scoring_code_path, source_code=source_code)

        self.log.info(
            f"Scoring code for project_id={self.project_id} and model_id={self.model_id} "
            f"downloaded to the filepath={scoring_code_path}"
        )

        return scoring_code_path



# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.

from airflow.decorators import dag
from airflow.operators.empty import EmptyOperator
from datarobot.enums import DataWranglingDialect

from datarobot_provider.example_dags.wrangler_example_recipe import WRANGLER_EXAMPLE_RECIPE
from datarobot_provider.operators.data_prep import CreateWranglingRecipeOperator
from datarobot_provider.operators.data_registry import CreateDatasetFromRecipeOperator
from datarobot_provider.operators.data_registry import UploadDatasetOperator
from datarobot_provider.operators.datarobot import GetOrCreateUseCaseOperator
from datarobot_provider.operators.deployment import DeployRegisteredModelOperator
from datarobot_provider.operators.deployment import ScorePredictionsOperator
from datarobot_provider.operators.monitoring import UpdateDriftTrackingOperator
from datarobot_provider.sensors.datarobot import ScoringCompleteSensor

"""
Example of Aiflow DAG for DataRobot data deployment and prediction generation.
Configurable parameters for this dag:
* model_package_id - The ID of the DataRobot model package (version) to deploy.
* deployment_label - A human readable label of the deployment.
* default_prediction_server_id - an identifier of a prediction server to be used as the default prediction server
  When working with prediction environments, default prediction server Id should not be provided
* target_drift_enabled - if target drift tracking is to be turned on
* feature_drift_enabled - if feature drift tracking is to be turned on
* predictions_dataset_file_path - the path to the dataset to be used for predictions

PREDICTION SERVER IDS:
See https://docs.datarobot.com/en/docs/api/reference/predapi/pred-server-id.html

"""


@dag(
    schedule=None,
    tags=["example", "csv", "predictions", "deployment"],
    params={
        "model_package_id": "",
        "deployment_label": "hospital-readmissions-example-deployment-prediction",
        "default_prediction_server_id": "",
        "target_drift_enabled": True,
        "feature_drift_enabled": True,
        "predictions_dataset_file_path": "https://s3.amazonaws.com/datarobot_public_datasets/10k_diabetes.csv",
    },
)
def deployment_prediction_generation():
    # Create a Use Case to keep all subsequent assets. Default name is "Airflow"
    create_use_case = GetOrCreateUseCaseOperator(task_id="create_use_case", set_default=True)

    deploy_registered_model = DeployRegisteredModelOperator(
        task_id="deploy_registered_model",
        model_package_id="{{ params.model_package_id }}",
        deployment_label="{{ params.deployment_label }}",
        extra_params={"default_prediction_server_id": "{{ params.default_prediction_server_id }}"},
    )

    update_drift_tracking = UpdateDriftTrackingOperator(
        task_id="update_drift_tracking",
        deployment_id=str(deploy_registered_model.output),
        target_drift_enabled=bool("{{ params.target_drift_enabled }}"),
        feature_drift_enabled=bool("{{ params.feature_drift_enabled }}"),
    )

    # Upload the data into Data Registry.
    predictions_dataset = UploadDatasetOperator(
        task_id="upload_dataset", file_path="{{ params.predictions_dataset_file_path }}"
    )

    # Define data preparation.
    predictions_recipe = CreateWranglingRecipeOperator(
        task_id="create_recipe",
        dataset_id=str(predictions_dataset.output),
        dialect=DataWranglingDialect.SPARK,
        operations=WRANGLER_EXAMPLE_RECIPE,
    )

    # Apply data preparation and save the modified data in the Data Registry.
    publish_recipe = CreateDatasetFromRecipeOperator(
        task_id="publish_recipe",
        recipe_id=str(predictions_recipe.output),
        do_snapshot=True,
    )

    deployment_predictions = ScorePredictionsOperator(
        task_id="deployment_predictions",
        deployment_id=str(deploy_registered_model.output),
        score_settings={
            "intake_settings": {
                "type": "dataset",
                "dataset_id": str(publish_recipe.output),
            }
        },
    )

    scoring_complete_sensor = ScoringCompleteSensor(
        task_id="check_scoring_complete",
        job_id=str(deployment_predictions.output),
    )

    collect_ops = EmptyOperator(task_id="collect_ops")

    (create_use_case >> [deploy_registered_model, predictions_dataset])
    (deploy_registered_model >> update_drift_tracking)
    (predictions_dataset >> predictions_recipe >> publish_recipe)
    ([update_drift_tracking, publish_recipe] >> collect_ops)
    (collect_ops >> deployment_predictions >> scoring_complete_sensor)


deployment_prediction_generation()



# Copyright 2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from collections.abc import Sequence
from typing import Any
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import datarobot as dr
from airflow.exceptions import AirflowFailException
from airflow.utils.context import Context
from datarobot import Blueprint
from strenum import StrEnum

from datarobot_provider.operators.base_datarobot_operator import XCOM_DEFAULT_USE_CASE_ID
from datarobot_provider.operators.base_datarobot_operator import BaseDatarobotOperator
from datarobot_provider.operators.base_datarobot_operator import BaseUseCaseEntityOperator

DATAROBOT_MAX_WAIT = 3600
DATAROBOT_AUTOPILOT_TIMEOUT = 86400
DATETIME_FORMAT = "%Y-%m-%d %H:%M:%s"


class GetOrCreateUseCaseOperator(BaseDatarobotOperator):
    """
    Creates a DataRobot Use Case.

    Args:
        datarobot_conn_id (str): Connection ID, defaults to `datarobot_default`.
        name (str): Use Case name.
        description (Optional[str]): Use Case description.
        reuse_policy (CreateUseCaseOperator.ReusePolicy): Should the operator reuse an existing Use Case with the same *name*?

            EXACT: Reuse the Use Case if it has exactly the same *name* and *description*.
            SEARCH_BY_NAME_UPDATE_DESCRIPTION: Reuse the Use Case if it has exactly the same *name*. Update *description* if it's different.
            SEARCH_BY_NAME_PRESERVE_DESCRIPTION: Reuse the Use Case if it has exactly the same *name*. Don't modify *description*.
            NO_REUSE: Always create a new Use Case.

            default: EXACT.
        set_default (bool): Set this Use Case as a default one for all subsequent tasks in the DAG.

    Returns:
        str: DataRobot UseCase ID.
    """

    class ReusePolicy(StrEnum):
        EXACT = "EXACT"
        SEARCH_BY_NAME_UPDATE_DESCRIPTION = "SEARCH_BY_NAME_UPDATE_DESCRIPTION"
        SEARCH_BY_NAME_PRESERVE_DESCRIPTION = "SEARCH_BY_NAME_PRESERVE_DESCRIPTION"
        NO_REUSE = "NO_REUSE"

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = ["name", "description"]

    def __init__(
        self,
        *,
        name: str = "{{ params.use_case_name | default('Airflow') }}",
        description: Optional[str] = "{{ params.use_case_description | default('') }}",
        reuse_policy: ReusePolicy = ReusePolicy.EXACT,
        set_default: bool = False,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.name = name
        self.description = description
        self.reuse_policy = reuse_policy
        self.set_default = set_default

    def execute(self, context: Context) -> Optional[str]:
        use_case = None

        if self.reuse_policy != self.ReusePolicy.NO_REUSE:
            use_case = self._search_for_existing_use_case()

            if (
                use_case is not None
                and self.reuse_policy == self.ReusePolicy.SEARCH_BY_NAME_UPDATE_DESCRIPTION
                and use_case.description != self.description
            ):
                self.log.info(
                    'Update Use Case description from "%s" to "%s"',
                    use_case.description,
                    self.description,
                )
                use_case.update(self.name, self.description)

        if use_case is None:
            self.log.info("Creating DataRobot Use Case")
            use_case = dr.UseCase.create(name=self.name, description=self.description)
            self.log.info(f"Use case created: use_case_id={use_case.id}")

        if self.set_default:
            self.log.info(
                'Set "%(name)s" (id=%(use_case_id)s) as a default Use Case.',
                {"name": use_case.name, "use_case_id": use_case.id},
            )
            self.xcom_push(context, XCOM_DEFAULT_USE_CASE_ID, use_case.id)

        return use_case.id

    def _search_for_existing_use_case(self) -> Optional[dr.UseCase]:
        candidates = []

        for use_case in dr.UseCase.list(search_params={"search": self.name}):
            if use_case.name == self.name:
                if (
                    self.reuse_policy == self.ReusePolicy.EXACT
                    and self.description
                    and use_case.description != self.description
                ):
                    continue

                self.log.info("Use an existing Use Case id=%s", use_case.id)

                candidates.append(use_case)

        if len(candidates) == 0:
            return None

        if (
            len(candidates) > 1
            and self.description
            and any(x.description == self.description for x in candidates)
        ):
            candidates = [x for x in candidates if x.description == self.description]

        return max(candidates, key=lambda x: x.created_at)


class CreateProjectOperator(BaseUseCaseEntityOperator):
    """
    Creates DataRobot project.

    Args:
        dataset_id (str, optional): DataRobot AI Catalog dataset ID.
        dataset_version_id (str, optional): DataRobot AI Catalog dataset version ID.
        datarobot_conn_id (str, optional): Connection ID, defaults to `datarobot_default`.
        recipe_id (str, optional): DataRobot Recipe ID.

    Returns:
        str: DataRobot project ID.
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = [
        "dataset_id",
        "dataset_version_id",
        "credential_id",
        "recipe_id",
        "use_case_id",
    ]

    def __init__(
        self,
        *,
        dataset_id: Optional[str] = None,
        dataset_version_id: Optional[str] = None,
        credential_id: Optional[str] = None,
        recipe_id: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.dataset_id = dataset_id
        self.dataset_version_id = dataset_version_id
        self.credential_id = credential_id
        self.recipe_id = recipe_id

    def execute(self, context: Context) -> Optional[str]:
        use_case = self.get_use_case(context)

        # Create DataRobot project
        self.log.info("Creating DataRobot project")

        if self.dataset_id is None and "training_data" in context["params"]:
            # training_data may be a pre-signed URL to a file on S3 or a path to a local file
            project: dr.Project = dr.Project.create(
                context["params"]["training_data"],
                context["params"]["project_name"],
                use_case=use_case,
            )
            self.log.info(f"Project created: project_id={project.id} from local file")
            project.unsupervised_mode = context["params"].get("unsupervised_mode")
            project.unlock_holdout()
            return project.id
        elif self.dataset_id is not None or "training_dataset_id" in context["params"]:
            # training_dataset_id may be provided via params
            # or dataset_id should be returned from previous operator
            training_dataset_id = (
                self.dataset_id
                if self.dataset_id is not None
                else context["params"]["training_dataset_id"]
            )

            project = dr.Project.create_from_dataset(
                dataset_id=training_dataset_id,
                dataset_version_id=self.dataset_version_id,
                credential_id=self.credential_id,
                project_name=context["params"]["project_name"],
                use_case=use_case,
            )
            # Some weird problem with mypy: it passes here locally, but fails in CI
            self.log.info(
                f"Project created: project_id={project.id} from dataset: dataset_id={training_dataset_id}"  # type: ignore[attr-defined, unused-ignore]
            )
            return project.id  # type: ignore[attr-defined, unused-ignore]

        elif self.recipe_id is not None:
            project = dr.Project.create_from_recipe(
                recipe_id=self.recipe_id,
                use_case=use_case,
            )
            self.log.info(
                f"Project created: project_id={project.id} from recipe: recipe_id={self.recipe_id}"  # type: ignore[attr-defined, unused-ignore]
            )
            return project.id  # type: ignore[attr-defined, unused-ignore]

        else:
            raise AirflowFailException(
                "For Project creation one of training_data, training_dataset_id or "
                "recipe_id must be provided"
            )


class TrainModelsOperator(BaseDatarobotOperator):
    """
    Triggers DataRobot Autopilot to train models.

    Args:
        project_id (str): DataRobot project ID.
        datarobot_conn_id (str, optional): Connection ID, defaults to `datarobot_default`.
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = ["project_id"]

    def __init__(
        self,
        *,
        project_id: str,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id

    def execute(self, context: Context) -> None:
        # Train models
        project = dr.Project.get(self.project_id)
        if project.target:
            self.log.info(f"Models are already trained for project_id={project.id}")
        else:
            self.log.info(
                f"Starting DataRobot Autopilot for project_id={project.id} "
                f"with settings={context['params']['autopilot_settings']}"
            )
            project.analyze_and_model(**context["params"]["autopilot_settings"])


class SelectBestModelOperator(BaseDatarobotOperator):
    """
    Selects the best model from a DataRobot project based on a specified evaluation metric.

    If the evaluation metric is not provided, the operator uses the project's primary metric.

    Args:
        project_id (str): DataRobot project ID.
        metric (str, optional): The evaluation metric used to rank models.

    Returns:
        str: The best model's ID as a string.
    """

    template_fields = ["project_id"]

    def __init__(self, *, project_id: str, metric: Optional[str] = None, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.metric = metric

    def validate(self) -> None:
        if not self.project_id:
            raise AirflowFailException("The `project_id` parameter is required.")

    def execute(self, context: Context) -> str:
        self.log.info(f"Selecting top model for project_id: {self.project_id}")
        project = dr.Project.get(self.project_id)

        if not self.metric:
            self.metric = project.metric
        best_model = project.get_top_model(metric=self.metric)
        return str(best_model.id)


class GetProjectModelsOperator(BaseDatarobotOperator):
    """
    Returns a list of all trained models and their ids.

    Args:
        project_id (str): DataRobot project ID.

    Returns:
        List[Tuple[str, str]]: The best model's ID as a string.
    """

    template_fields = ["project_id"]

    def __init__(self, *, project_id: str, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id

    def validate(self) -> None:
        if not self.project_id:
            raise AirflowFailException("The `project_id` parameter is required.")

    def execute(self, context: Context) -> List[Tuple[str, str]]:
        self.log.info(f"Selecting top model for project_id: {self.project_id}")
        project = dr.Project.get(self.project_id)
        models = project.get_models()

        return [(str(model.id), str(model.model_type)) for model in models]


class GetProjectBlueprintsOperator(BaseDatarobotOperator):
    """
    Get DataRobot project blueprints. Blueprint ids can optionally
    be filtered by a specific model type.

    Args:
        project_id (str): DataRobot project ID.
        return_all (bool, optional): Return all blueprint ids, default is False.
        filter_model_type (str, optional): Blueprint filter type, default returns all ids.
        datarobot_conn_id (str, optional): Connection ID, defaults to `datarobot_default`.

    Returns:
        list of str: List of DataRobot blueprint IDs.
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = ["project_id"]

    def __init__(
        self,
        *,
        project_id: str,
        return_all: Optional[bool] = False,
        filter_model_type: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.filter_model_type = filter_model_type
        self.return_all = return_all

    def execute(self, context: Context) -> Union[List[Union[str, None]], str]:
        # Get DataRobot project blueprints
        project = dr.Project.get(self.project_id)
        blueprints: List[Blueprint] = project.get_blueprints()
        if self.filter_model_type:
            # Filter blueprints by type
            blueprint_ids = [
                blueprint.id
                for blueprint in blueprints
                if self.filter_model_type in str(blueprint.model_type).lower()
            ]
        else:
            blueprint_ids = [blueprint.id for blueprint in blueprints]

        if blueprint_ids is None:
            raise AirflowFailException("No blueprints found for project_id={self.project_id}")

        if self.return_all:
            return blueprint_ids
        else:
            return str(blueprint_ids[0])



# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from datetime import datetime

from airflow.decorators import dag
from airflow.decorators import task

from datarobot_provider.operators.monitoring import GetAccuracyOperator
from datarobot_provider.operators.monitoring import GetServiceStatsOperator


@dag(
    schedule=None,
    start_date=datetime(2023, 1, 1),
    tags=["example", "mlops"],
)
def deployment_service_stats_and_accuracy(deployment_id=None):
    if not deployment_id:
        raise ValueError("Invalid or missing `deployment_id` value")

    service_stats_op = GetServiceStatsOperator(
        task_id="get_service_stats",
        deployment_id=deployment_id,
    )

    get_accuracy_op = GetAccuracyOperator(task_id="get_accuracy", deployment_id=deployment_id)

    @task(task_id="example_accuracy_check_python")
    def example_metrics_processing(model_service_stat, model_accuracy):
        """Example of custom logic based on metrics from the deployment."""

        # Put your service stat processing logic here:
        current_model_id = model_accuracy["model_id"]

        total_predictions = model_service_stat["metrics"]["totalPredictions"]
        print(f"model_id:{current_model_id}, total_predictions:{total_predictions}")
        print(f"model_id:{current_model_id}, model_accuracy: {model_accuracy}")

        return total_predictions

    metrics_processing = example_metrics_processing(
        model_service_stat=service_stats_op.output, model_accuracy=get_accuracy_op.output
    )

    (service_stats_op, get_accuracy_op) >> metrics_processing


deployment_metrics_dag = deployment_service_stats_and_accuracy()

if __name__ == "__main__":
    deployment_metrics_dag.test()



# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
"""
Config example for this dag:
{
    "actual_value_column": 'ACTUAL',
    "association_id_column": 'id',
    "timestamp_column": 'timestamp',
    "was_acted_on_column": 'acted',
}
"""

from datetime import datetime

from airflow.decorators import dag

from datarobot_provider.operators.mlops import SubmitActualsFromCatalogOperator
from datarobot_provider.sensors.client import BaseAsyncResolutionSensor


@dag(
    schedule_interval=None,
    start_date=datetime(2023, 1, 1),
    tags=["example"],
    # Default json config example:
    params={
        "association_id_column": "id",  # column name with a unique identifier used with a prediction.
        "actual_value_column": "ACTUAL",  # column name with the actual value of a prediction.
        "timestamp_column": "",  # optional - column name with datetime or string in RFC3339 format.
        "was_acted_on_column": "",  # optional
    },
)
def datarobot_submit_actuals_from_ai_catalog(deployment_id=None, dataset_id=None):
    if not deployment_id:
        raise ValueError("Invalid or missing `deployment_id` value")
    if not dataset_id:
        raise ValueError("Invalid or missing `dataset_id` value")

    upload_actuals_op = SubmitActualsFromCatalogOperator(
        task_id="upload_actuals",
        deployment_id=deployment_id,
        # dataset_id can be received from previous operator
        dataset_id=dataset_id,
    )

    uploading_complete_sensor = BaseAsyncResolutionSensor(
        task_id="check_uploading_actuals_complete",
        job_id=upload_actuals_op.output,
        poke_interval=5,
        mode="reschedule",
        timeout=3600,
    )

    upload_actuals_op >> uploading_complete_sensor


datarobot_upload_actuals_from_ai_catalog_dag = datarobot_submit_actuals_from_ai_catalog()

if __name__ == "__main__":
    datarobot_upload_actuals_from_ai_catalog_dag.test()



# Copyright 2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from datetime import datetime

from airflow.decorators import dag
from airflow.decorators import task

from datarobot_provider.operators.bias_and_fairness import UpdateBiasAndFairnessSettingsOperator
from datarobot_provider.operators.data_registry import CreateDatasetFromDataStoreOperator
from datarobot_provider.operators.data_registry import GetDataStoreOperator
from datarobot_provider.operators.datarobot import CreateProjectOperator
from datarobot_provider.operators.datarobot import TrainModelsOperator
from datarobot_provider.operators.deployment import DeployRecommendedModelOperator
from datarobot_provider.operators.deployment import GetFeatureDriftOperator
from datarobot_provider.operators.deployment import GetTargetDriftOperator
from datarobot_provider.operators.deployment import ScorePredictionsOperator
from datarobot_provider.operators.monitoring import GetAccuracyOperator
from datarobot_provider.operators.monitoring import GetServiceStatsOperator
from datarobot_provider.operators.monitoring import UpdateMonitoringSettingsOperator
from datarobot_provider.operators.monitoring_job import BatchMonitoringOperator
from datarobot_provider.operators.segment_analysis import UpdateSegmentAnalysisSettingsOperator
from datarobot_provider.sensors.datarobot import AutopilotCompleteSensor
from datarobot_provider.sensors.datarobot import ScoringCompleteSensor
from datarobot_provider.sensors.monitoring_job import MonitoringJobCompleteSensor


@dag(
    schedule=None,
    start_date=datetime(2023, 1, 1),
    tags=["example", "end-to-end", "pipline"],
    params={
        "data_connection": "Demo Connection",
        "dataset_name": "Demo-Airflow-training-dataset",
        "table_schema": "DEMO_SCHEMA",
        "table_name": "DEMO-TRAINING-DATASET",
        "persist_data_after_ingestion": False,
        "do_snapshot": False,
        "project_name": "Airflow - Demo",
        "unsupervised_mode": False,
        "use_feature_discovery": False,
        "autopilot_settings": {"target": "readmitted", "mode": "quick", "max_wait": 3600},
        "deployment_label": "Deployment from Airflow - Demo",
        "deployment_description": "Airflow - Demo Deployment",
        "score_settings": {
            "intake_settings": {
                "type": "jdbc",
                "table": "DEMO-SCORING-DATASET",
                "schema": "DEMO_SCHEMA",
            },
            "output_settings": {
                "type": "jdbc",
                "schema": "DEMO_SCHEMA",
                "table": "DEMO-SCORING-OUTPUT-DATASET",
                "statement_type": "insert",
                "create_table_if_not_exists": True,
            },
            # "passthrough_columns": ['column1', 'column2'],
            "passthrough_columns_set": "all",
        },
        "segment_analysis_enabled": True,
        "segment_analysis_attributes": ["race", "gender"],
        "protected_features": ["gender"],
        "preferable_target_value": "True",
        "fairness_metric_set": "equalParity",
        "fairness_threshold": 0.1,
        "target_drift_enabled": True,
        "feature_drift_enabled": True,
        "association_id_column": ["rowID"],
        "required_association_id": True,
        "predictions_data_collection_enabled": True,
        "monitoring_settings": {
            "intake_settings": {
                "type": "jdbc",
                "table": "DEMO-ACTUALS-DATASET",  # table with actuals
                "schema": "DEMO_SCHEMA",
            },
            "monitoring_columns": {
                "association_id_column": "rowID",
                "actuals_value_column": "ACTUALS",
            },
        },
    },
)
def advanced_datarobot_pipeline_jdbc():
    get_jdbc_connection_op = GetDataStoreOperator(task_id="get_jdbc_connection")

    dataset_connect_op = CreateDatasetFromDataStoreOperator(
        task_id="create_dataset_jdbc",
        data_store_id=get_jdbc_connection_op.output,
    )

    create_project_op = CreateProjectOperator(
        task_id="create_project",
        dataset_id=dataset_connect_op.output,
    )

    train_models_op = TrainModelsOperator(
        task_id="train_models",
        project_id=create_project_op.output,
    )

    autopilot_complete_sensor = AutopilotCompleteSensor(
        task_id="check_autopilot_complete",
        project_id=create_project_op.output,
    )

    deploy_model_op = DeployRecommendedModelOperator(
        task_id="deploy_recommended_model",
        project_id=create_project_op.output,
    )

    update_segment_analysis_settings_op = UpdateSegmentAnalysisSettingsOperator(
        task_id="update_segment_analysis_settings",
        deployment_id=deploy_model_op.output,
    )

    update_bias_and_fairness_settings_op = UpdateBiasAndFairnessSettingsOperator(
        task_id="update_bias_and_fairness_settings",
        deployment_id=deploy_model_op.output,
    )

    update_monitoring_settings_op = UpdateMonitoringSettingsOperator(
        task_id="update_monitoring_settings",
        deployment_id=deploy_model_op.output,
    )

    score_predictions_op = ScorePredictionsOperator(
        task_id="score_predictions_jdbc",
        deployment_id=deploy_model_op.output,
        intake_datastore_id=get_jdbc_connection_op.output,
        output_datastore_id=get_jdbc_connection_op.output,
    )

    scoring_complete_sensor = ScoringCompleteSensor(
        task_id="check_scoring_complete",
        job_id=score_predictions_op.output,
    )

    service_stats_op = GetServiceStatsOperator(
        task_id="get_service_stats",
        deployment_id=deploy_model_op.output,
    )

    target_drift_op = GetTargetDriftOperator(
        task_id="target_drift",
        deployment_id=deploy_model_op.output,
    )

    feature_drift_op = GetFeatureDriftOperator(
        task_id="feature_drift",
        deployment_id=deploy_model_op.output,
    )

    batch_monitoring_op = BatchMonitoringOperator(
        task_id="batch_monitoring",
        deployment_id=deploy_model_op.output,
        datastore_id=get_jdbc_connection_op.output,
    )

    batch_monitoring_complete_sensor = MonitoringJobCompleteSensor(
        task_id="check_monitoring_job_complete",
        job_id=batch_monitoring_op.output,
        poke_interval=5,
        mode="reschedule",
        timeout=3600,
    )

    get_accuracy_op = GetAccuracyOperator(
        task_id="get_accuracy", deployment_id=deploy_model_op.output
    )

    @task(task_id="example_processing_python")
    def service_stat_processing(model_service_stat, feature_drift, target_drift, accuracy):
        """Example of custom logic based on service stats from the deployment."""

        # Put your service stat processing logic here:
        current_model_id = model_service_stat["model_id"]
        total_predictions = model_service_stat["metrics"]["totalPredictions"]
        # example of custom logic:
        return {"model_id": current_model_id, "alive": total_predictions > 0}

    example_service_stat_processing = service_stat_processing(
        model_service_stat=service_stats_op.output,
        feature_drift=target_drift_op.output,
        target_drift=feature_drift_op.output,
        accuracy=get_accuracy_op.output,
    )

    (
        get_jdbc_connection_op
        >> dataset_connect_op
        >> create_project_op
        >> train_models_op
        >> autopilot_complete_sensor
        >> deploy_model_op
        >> update_monitoring_settings_op
        >> update_segment_analysis_settings_op
        >> update_bias_and_fairness_settings_op
        >> score_predictions_op
        >> scoring_complete_sensor
        >> batch_monitoring_op
        >> batch_monitoring_complete_sensor
        >> (get_accuracy_op, target_drift_op, feature_drift_op, service_stats_op)
        >> example_service_stat_processing
    )


advanced_datarobot_pipeline_jdbc_dag = advanced_datarobot_pipeline_jdbc()
if __name__ == "__main__":
    advanced_datarobot_pipeline_jdbc_dag.test()



# Copyright 2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from collections.abc import Sequence
from typing import Any
from typing import Optional

import datarobot as dr
from airflow.utils.context import Context

from datarobot_provider.operators.base_datarobot_operator import BaseDatarobotOperator

DEFAULT_MAX_WAIT_SEC = 600


class PredictionExplanationsInitializationOperator(BaseDatarobotOperator):
    """
    Triggering a prediction explanations initialization of a model.
    :param project_id: DataRobot project ID
    :type project_id: str
    :param model_id: DataRobot model ID
    :type model_id: str
    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    :return: Prediction Explanations Initialization job ID
    :rtype: str
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = [
        "project_id",
        "model_id",
    ]

    def __init__(
        self,
        *,
        project_id: Optional[str] = None,
        model_id: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.model_id = model_id

    def validate(self) -> None:
        if self.project_id is None:
            raise ValueError(
                "project_id is required to trigger a prediction explanations initialization."
            )

        if self.model_id is None:
            raise ValueError(
                "model_id is required to trigger a prediction explanations initialization."
            )

    def execute(self, context: Context) -> str:
        # Initialize prediction explanations
        pei_job = dr.PredictionExplanationsInitialization.create(self.project_id, self.model_id)

        self.log.info(
            f"Triggered prediction explanations initialization of a model, job_id={pei_job.id}"
        )

        return pei_job.id


class ComputePredictionExplanationsOperator(BaseDatarobotOperator):
    """
    Create prediction explanations for the specified dataset.
    :param project_id: DataRobot project ID
    :type project_id: str
    :param model_id: DataRobot model ID
    :type model_id: str
    :param external_dataset_id: DataRobot external dataset ID
    :type external_dataset_id: str
    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    :return: Prediction Explanations Initialization job ID
    :rtype: str
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = [
        "project_id",
        "model_id",
        "external_dataset_id",
    ]

    def __init__(
        self,
        *,
        project_id: Optional[str] = None,
        model_id: Optional[str] = None,
        external_dataset_id: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.model_id = model_id
        self.external_dataset_id = external_dataset_id

    def validate(self) -> None:
        if self.project_id is None:
            raise ValueError("project_id is required to compute prediction explanations.")

        if self.model_id is None:
            raise ValueError("model_id is required to compute prediction explanations.")

        if self.external_dataset_id is None:
            raise ValueError("external_dataset_id is required to compute prediction explanations.")

    def execute(self, context: Context) -> str:
        # Creating compute prediction explanations job:
        pe_job = dr.PredictionExplanations.create(
            self.project_id,
            self.model_id,
            self.external_dataset_id,
            max_explanations=context["params"].get("max_explanations", None),
            threshold_low=context["params"].get("threshold_low", None),
            threshold_high=context["params"].get("threshold_high", None),
        )

        self.log.info(
            f"Triggered prediction explanations for the specified dataset, job_id={pe_job.id}"
        )

        return pe_job.id



# Copyright 2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.

from typing import Union

from airflow.sensors.base import PokeReturnValue
from airflow.utils.context import Context
from datarobot import ModelJob

from datarobot_provider.sensors.model_insights import DataRobotJobSensor


class ModelTrainingJobSensor(DataRobotJobSensor):
    """
    Checks whether DataRobot Model Training Job is complete.

    Args:
        project_id (str): DataRobot project ID.
        job_id (str): DataRobot Job ID.
        datarobot_conn_id (str, optional): Connection ID, defaults to `datarobot_default`.

    Returns:
        bool | PokeReturnValue: False if not yet completed, PokeReturnValue(True, trained_model.id) if model training completed.
    """

    def get_job_result(self, context: Context) -> Union[bool, PokeReturnValue]:
        trained_model = ModelJob.get_model(self.project_id, self.job_id)
        return PokeReturnValue(True, trained_model.id)



# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from collections.abc import Sequence
from typing import Any
from typing import Optional

import datarobot as dr
from airflow.utils.context import Context

from datarobot_provider.operators.base_datarobot_operator import BaseDatarobotOperator

DATETIME_FORMAT = "%Y-%m-%d %H:%M:%s"


class GetServiceStatsOperator(BaseDatarobotOperator):
    """
    Gets service stats measurements from a deployment.

    :param deployment_id: DataRobot deployment ID
    :type deployment_id: str
    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    :return: Service stats for a Deployment
    :rtype: list[dict]
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = ["deployment_id"]

    def __init__(
        self,
        *,
        deployment_id: str,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.deployment_id = deployment_id

    def execute(self, context: Context) -> list[dict]:
        self.log.info(f"Getting service stats for deployment_id={self.deployment_id}")
        deployment = dr.Deployment.get(self.deployment_id)
        service_stats_params = context["params"].get("service_stats", {})
        service_stats = deployment.get_service_stats(**service_stats_params)
        return _serialize_metrics(service_stats)


def _serialize_metrics(service_stats_obj, date_format=DATETIME_FORMAT):
    service_stats_dict = service_stats_obj.__dict__.copy()
    service_stats_dict["period"] = {
        "start": service_stats_obj.period["start"].strftime(date_format),
        "end": service_stats_obj.period["end"].strftime(date_format),
    }
    return service_stats_dict


class GetAccuracyOperator(BaseDatarobotOperator):
    """
    Gets the accuracy of a deployment’s predictions.

    :param deployment_id: DataRobot deployment ID
    :type deployment_id: str
    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    :return: accuracy for a Deployment
    :rtype: list[dict]
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = ["deployment_id"]

    def __init__(
        self,
        *,
        deployment_id: str,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.deployment_id = deployment_id

    def execute(self, context: Context) -> list[dict]:
        self.log.info(f"Getting service stats for deployment_id={self.deployment_id}")
        deployment = dr.Deployment.get(self.deployment_id)
        accuracy_params = context["params"].get("accuracy", {})
        accuracy = deployment.get_accuracy(**accuracy_params)
        return _serialize_metrics(accuracy)


class GetMonitoringSettingsOperator(BaseDatarobotOperator):
    """
    Get monitoring settings for deployment.

    :param deployment_id: DataRobot deployment ID
    :type deployment_id: str
    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = ["deployment_id"]

    def __init__(
        self,
        *,
        deployment_id: str,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.deployment_id = deployment_id

    def execute(self, context: Context) -> dict:
        self.log.info(f"Get Deployment for deployment_id={self.deployment_id}")
        deployment = dr.Deployment.get(deployment_id=self.deployment_id)

        drift_tracking_settings = deployment.get_drift_tracking_settings()
        association_id_settings = deployment.get_association_id_settings()
        predictions_data_collection_settings = deployment.get_predictions_data_collection_settings()

        monitoring_settings = {
            "drift_tracking_settings": drift_tracking_settings,
            "association_id_settings": association_id_settings,
            "predictions_data_collection_settings": predictions_data_collection_settings,
        }

        return monitoring_settings


class UpdateMonitoringSettingsOperator(BaseDatarobotOperator):
    """
    Updates monitoring settings for a deployment.

    :param deployment_id: DataRobot deployment ID
    :type deployment_id: str
    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = ["deployment_id"]

    def __init__(
        self,
        *,
        deployment_id: str,
        monitoring_settings: Optional[dict] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.deployment_id = deployment_id
        self.monitoring_settings = monitoring_settings

    def execute(self, context: Context) -> None:
        self.log.info(f"Getting Deployment for deployment_id={self.deployment_id}")
        deployment = dr.Deployment.get(deployment_id=self.deployment_id)

        current_drift_tracking_settings = deployment.get_drift_tracking_settings()

        target_drift_enabled = context["params"].get(
            "target_drift_enabled", current_drift_tracking_settings["target_drift"]["enabled"]
        )
        feature_drift_enabled = context["params"].get(
            "feature_drift_enabled", current_drift_tracking_settings["feature_drift"]["enabled"]
        )

        if (target_drift_enabled != current_drift_tracking_settings["target_drift"]["enabled"]) or (
            feature_drift_enabled != current_drift_tracking_settings["feature_drift"]["enabled"]
        ):
            self.log.debug(
                f"Trying to update drift settings for deployment_id={self.deployment_id}"
            )
            deployment.update_drift_tracking_settings(
                target_drift_enabled=target_drift_enabled,
                feature_drift_enabled=feature_drift_enabled,
            )
            self.log.info(
                f"Deployment drift settings updated for deployment_id={self.deployment_id}"
            )
        else:
            self.log.info(
                f"No need to update drift settings for deployment_id={self.deployment_id}"
            )

        # Possibly a bug: deployment.get_association_id_settings returns str
        current_association_id_settings: dict = deployment.get_association_id_settings()  # type: ignore[assignment]

        association_id_column = context["params"].get(
            "association_id_column", current_association_id_settings["column_names"]
        )
        required_in_prediction_requests = context["params"].get(
            "required_association_id",
            current_association_id_settings["required_in_prediction_requests"],
        )

        if (association_id_column != current_association_id_settings["column_names"]) or (
            required_in_prediction_requests
            != current_association_id_settings["required_in_prediction_requests"]
        ):
            deployment.update_association_id_settings(
                column_names=association_id_column,
                required_in_prediction_requests=required_in_prediction_requests,
            )
            self.log.info(
                f"Deployment association_id settings updated for deployment_id={self.deployment_id}"
            )
        else:
            self.log.info(
                f"No need to update association_id settings for deployment_id={self.deployment_id}"
            )

        current_predictions_data_collection_settings = (
            deployment.get_predictions_data_collection_settings()
        )
        predictions_data_collection_settings = context["params"].get(
            "predictions_data_collection_enabled",
            current_predictions_data_collection_settings["enabled"],
        )
        if (
            predictions_data_collection_settings
            != current_predictions_data_collection_settings["enabled"]
        ):
            deployment.update_predictions_data_collection_settings(
                enabled=predictions_data_collection_settings,
            )
            self.log.info(
                f"Deployment predictions data collection settings updated for deployment_id={self.deployment_id}"
            )
        else:
            self.log.info(
                f"No need to update predictions data collection settings for deployment_id={self.deployment_id}"
            )


class UpdateDriftTrackingOperator(BaseDatarobotOperator):
    """
    Update drift tracking settings for a DataRobot deployment using DataRobot's API.

    This operator updates drift tracking settings for an existing deployment by calling
    the deployment's `update_drift_tracking_settings()` method. It allows optional extra parameters
    to be passed to the DataRobot client call.

    Args:
        deployment_id (str): The ID of the deployment to update.
        target_drift_enabled (bool): Boolean flag to enable target drift tracking.
        feature_drift_enabled (bool): Boolean flag to enable feature drift tracking.
        kwargs (dict): Additional keyword arguments passed to the BaseDatarobotOperator.
    """

    template_fields: Sequence[str] = [
        "deployment_id",
        "target_drift_enabled",
        "feature_drift_enabled",
    ]

    def __init__(
        self,
        *,
        deployment_id: str,
        target_drift_enabled: bool = True,
        feature_drift_enabled: bool = True,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.deployment_id = deployment_id
        self.target_drift_enabled = target_drift_enabled
        self.feature_drift_enabled = feature_drift_enabled

    def validate(self) -> None:
        if not self.deployment_id:
            raise ValueError("deployment_id must be provided.")

    def execute(self, context: Context) -> str:
        self.log.info("Updating drift tracking settings for deployment: %s", self.deployment_id)
        deployment = dr.Deployment(id=self.deployment_id)
        deployment.update_drift_tracking_settings(
            target_drift_enabled=self.target_drift_enabled,
            feature_drift_enabled=self.feature_drift_enabled,
        )
        self.log.info("Drift tracking settings updated for deployment: %s", self.deployment_id)
        return self.deployment_id



# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from collections.abc import Sequence
from typing import Any
from typing import Optional

import datarobot as dr
from airflow.utils.context import Context

from datarobot_provider.operators.base_datarobot_operator import BaseDatarobotOperator


class SubmitActualsFromCatalogOperator(BaseDatarobotOperator):
    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = [
        "deployment_id",
        "dataset_id",
        "dataset_version_id",
    ]

    def __init__(
        self,
        *,
        deployment_id: str,
        dataset_id: str,
        dataset_version_id: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.deployment_id = deployment_id
        self.dataset_id = dataset_id
        self.dataset_version_id = dataset_version_id

    def validate(self) -> None:
        if self.deployment_id is None:
            raise ValueError("deployment_id is required to submit actuals.")

        if self.dataset_id is None:
            raise ValueError("dataset_id is required to submit actuals.")

    def execute(self, context: Context) -> str:
        self.log.info("Uploading Actuals from AI Catalog")

        deployment = dr.Deployment.get(deployment_id=self.deployment_id)

        status_job = deployment.submit_actuals_from_catalog_async(
            dataset_id=self.dataset_id,
            dataset_version_id=self.dataset_version_id,
            actual_value_column=context["params"].get("actual_value_column", None),
            association_id_column=context["params"].get("association_id_column", None),
            timestamp_column=context["params"].get("timestamp_column", None),
            was_acted_on_column=context["params"].get("was_acted_on_column", None),
        )

        self.log.debug(f"Uploading Actuals from AI Catalog, job_id: {status_job.job_id}")

        return status_job.job_id



# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import json
from pathlib import Path
from typing import Any
from typing import Optional
from typing import TypedDict
from typing import Union

from airflow.exceptions import AirflowException
from airflow.utils.context import Context
from datarobot._experimental.models.notebooks import Notebook
from datarobot._experimental.models.notebooks.enums import ManualRunType
from datarobot._experimental.models.notebooks.session import StartSessionParameters

from datarobot_provider.operators.base_datarobot_operator import BaseDatarobotOperator


class NotebookParametersData(TypedDict):
    data: list[StartSessionParameters]


OperatorParametersType = Union[str, NotebookParametersData, None]


class NotebookRunOperator(BaseDatarobotOperator):
    """
    Runs a DataRobot Notebook.

    :param notebook_id: DataRobot notebook ID
    :type notebook_id: str
    :param notebook_path: Path to the notebook file. Must be provided if the notebook is part of a Codespace.
    :type notebook_path: str
    :param notebook_parameters: Parameters to be set as environment variables in the notebook session. Must be in the
        form of `{"data": [{"name": "FOO", "value": "MyValue"}, {"name": "BAR", "value": "OtherValue"}]}`
    :type notebook_parameters: dict, optional
    :return: The ID of the triggered notebook run.
    :rtype: str
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields = [
        "notebook_id",
        "notebook_path",
        "notebook_parameters",
    ]

    def __init__(
        self,
        *,
        notebook_id: str = "{{ params.notebook_id }}",
        notebook_path: Optional[str] = "{{ params.notebook_path | default('') }}",
        notebook_parameters: OperatorParametersType = "{{ params.notebook_parameters | default('') }}",
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.notebook_id = notebook_id
        self.notebook_path = notebook_path
        self.notebook_parameters = notebook_parameters
        self.parameters = None

    def _parse_notebook_parameters(self) -> None:
        if self.notebook_parameters:
            if isinstance(self.notebook_parameters, str):
                parsed_parameters = json.loads(self.notebook_parameters)
            else:
                parsed_parameters = self.notebook_parameters
            # DAGs using Airflow's `Param` model can't seem to take an array/list - it needs to be an object/dict
            self.parameters = parsed_parameters.get("data")

    def _validate_notebook_parameters(self) -> None:
        try:
            self._parse_notebook_parameters()
        except Exception as exc:
            self.log.warning(
                f"Error ({str(exc)}) parsing notebook parameters: {self.notebook_parameters}"
                f" type={type(self.notebook_parameters)}"
            )
            raise AirflowException("Please check the format of your notebook parameters.") from exc

    def _validate_notebook_path(self) -> None:
        path_suffix = ".ipynb"
        if self.notebook_path:
            path = Path(self.notebook_path)
            if not path.is_absolute():
                raise AirflowException(
                    f"Supplied notebook path ({self.notebook_path}) must be an absolute path."
                )
            if path.suffix != path_suffix:
                raise AirflowException(
                    f"Supplied notebook path ({self.notebook_path}) must end with '{path_suffix}."
                )

    def validate(self) -> None:
        self._validate_notebook_path()
        self._validate_notebook_parameters()

    def execute(self, context: Context) -> str:
        self._parse_notebook_parameters()

        # Fetch the notebook
        notebook = Notebook.get(notebook_id=self.notebook_id)

        if not notebook.use_case_id:
            raise AirflowException("Notebook should have use_case_id")

        # Run the notebook as a manual run of type "pipeline"
        manual_run = notebook.run(
            notebook_path=self.notebook_path,
            parameters=self.parameters,
            manual_run_type=ManualRunType.PIPELINE,
        )
        self.log.info(f"Notebook triggered. Manual run ID: {manual_run.id}")

        return manual_run.id



# Copyright 2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from typing import TypedDict

ConnectionType = TypedDict(
    "ConnectionType",
    {
        "hook-class-name": str,
        "connection-type": str,
    },
)


ProviderInfoType = TypedDict(
    "ProviderInfoType",
    {
        "package-name": str,
        "name": str,
        "description": str,
        "versions": list[str],
        "connection-types": list[ConnectionType],
        "extra-links": list[str],
    },
)


def get_provider_info() -> ProviderInfoType:
    return {
        "package-name": "airflow-provider-datarobot-early-access",
        "name": "DataRobot Airflow Provider",
        "description": "DataRobot Airflow provider.",
        "versions": ["0.4.0"],
        "connection-types": [
            {
                "hook-class-name": "datarobot_provider.hooks.datarobot.DataRobotHook",
                "connection-type": "http",
            }
        ],
        "extra-links": [],
    }



# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from datetime import datetime

from airflow.decorators import dag
from airflow.decorators import task

from datarobot_provider.operators.scoring_code import DownloadDeploymentScoringCodeOperator


@dag(
    schedule=None,
    start_date=datetime(2023, 1, 1),
    tags=["example", "scoring_code"],
    # Default json config example:
    params={
        "source_code": False,
        "include_agent": False,
        "include_prediction_explanations": False,
        "include_prediction_intervals": False,
        # as an option you can pass base_path here:
        # 'scoring_code_filepath': "/home/airflow/gcs/data/",
    },
)
def download_scoring_code(deployment_id=None):
    if not deployment_id:
        raise ValueError("Invalid or missing `deployment_id` value")

    download_deployment_scoring_code_op = DownloadDeploymentScoringCodeOperator(
        task_id="download_scoring_code_from_deployment",
        # you can pass deployment_id from previous operator here:
        deployment_id=deployment_id,
        # you can pass base_path from previous operator here:
        base_path="/.../models/",
    )

    @task(task_id="example_using_scoring_code")
    def using_scoring_code_example(scoring_code_filepath):
        """Example of using scoring code locally"""

        # Put your logic with the scoring code here:
        # Note that local scoring is not recommended to use because its burden an Airflow worker node.
        # The next scoring code example is only for demo and testing proposes,
        # should be used only with a very small datasets (not higher than a few megabytes).
        # To run this example, DataRobot Prediction Library (https://datarobot.github.io/datarobot-predict)
        # should be installed first.

        input_file = "/.../input_file.csv"
        output_file = "/.../output_predictions.csv"

        import pandas as pd
        from datarobot_predict.scoring_code import ScoringCodeModel

        model = ScoringCodeModel(scoring_code_filepath)
        input_df = pd.read_csv(input_file)
        output_df = model.predict(input_df)
        output_df.to_csv(output_file)

    example_using_scoring_code = using_scoring_code_example(
        scoring_code_filepath=download_deployment_scoring_code_op.output,
    )

    download_deployment_scoring_code_op >> example_using_scoring_code


download_scoring_code_dag = download_scoring_code()

if __name__ == "__main__":
    download_scoring_code_dag.test()



# Copyright 2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
"""
Config example for this dag:
{
    "deployment_id": "62cc0a6383d7a13d34f83344",
    "score_settings": {
        "intake_settings": {
            "type": "dataset",
            "dataset_id": "623d8ae79b186124a926c3cd"
        },
        "output_settings": {
            "type": "localFile",
            "path": "include/Diabetes_predictions.csv"
        }
    }
}
"""

from datetime import datetime

from airflow.decorators import dag

from datarobot_provider.operators.deployment import ScorePredictionsOperator
from datarobot_provider.sensors.datarobot import ScoringCompleteSensor


@dag(
    schedule=None,
    start_date=datetime(2022, 1, 1),
    tags=["example", "scoring"],
)
def datarobot_batch_scoring_templated():
    score_predictions_op = ScorePredictionsOperator(
        task_id="score_predictions",
        deployment_id="testdeploymentid",
        score_settings={
            "intake_settings": {"type": "dataset", "dataset_id": "testdatasetid"},
            "output_settings": {
                "type": "localFile",
                "path": "include/{{ ds_nodash }}/{{ params.myparam }}/Diabetes_predictions.csv",
            },
        },
        # custom parameter example
        params={"myparam": "test_param_value"},
    )

    scoring_complete_sensor = ScoringCompleteSensor(
        task_id="check_scoring_complete",
        job_id=score_predictions_op.output,
    )

    score_predictions_op >> scoring_complete_sensor


datarobot_batch_scoring_templated_dag = datarobot_batch_scoring_templated()

if __name__ == "__main__":
    datarobot_batch_scoring_templated_dag.test()



# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from collections.abc import Sequence
from typing import Any
from typing import List
from typing import Optional

import datarobot as dr
from airflow.utils.context import Context
from datarobot import QUEUE_STATUS

from datarobot_provider.operators.base_datarobot_operator import BaseDatarobotOperator


class GetJobsOperator(BaseDatarobotOperator):
    """
    Get a list of jobs that match the requested status.

    Args:
        project_id (str): DataRobot project ID.
        status (str): Status of the job. Default is QUEUE_STATUS.INPROGRESS.
        datarobot_conn_id (str, optional): Connection ID, defaults to `datarobot_default`.
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = ["project_id", "job_id"]

    def __init__(
        self,
        *,
        project_id: str,
        status: Optional[str] = QUEUE_STATUS.INPROGRESS,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.status = status

    def validate(self):
        if not self.project_id:
            raise ValueError("project_id is required.")
        if self.status not in list(QUEUE_STATUS.__dict__.values()):
            raise ValueError(f"Invalid status: {self.status}, must be a QUEUE_STATUS.")

    def execute(self, context: Context) -> List[str]:
        project = dr.Project.get(self.project_id)
        # Actual filtering logic is handled by the DataRobot API
        project_jobs = project.get_all_jobs(status=self.status)

        return [str(job.id) for job in project_jobs]


class CancelJobOperator(BaseDatarobotOperator):
    """
    Cancel a job that is in the queue or currently running.

    Args:
        project_id (str): DataRobot project ID.
        job_id (str): Job ID in the project.
        datarobot_conn_id (str, optional): Connection ID, defaults to `datarobot_default`.
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = ["project_id", "job_id"]

    def __init__(
        self,
        *,
        project_id: str,
        job_id: str,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.job_id = job_id

    def validate(self):
        if not self.project_id:
            raise ValueError("project_id is required.")
        if not self.job_id:
            raise ValueError("job_id is required.")

    def execute(self, context: Context) -> None:
        project = dr.Project.get(self.project_id)
        job = dr.Job.get(self.project_id, self.job_id)
        if job.status in [
            QUEUE_STATUS.INPROGRESS,
            QUEUE_STATUS.QUEUE,
        ]:
            project.pause_autopilot()
            job.cancel()
            project.unpause_autopilot()

        else:
            self.log.info("Job already completed or aborted.")



# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from enum import Enum
from typing import Any
from typing import Dict

import datarobot as dr
from airflow.exceptions import AirflowException
from airflow.utils.context import Context

from datarobot_provider.operators.base_datarobot_operator import BaseDatarobotOperator


class ModelType(Enum):
    LEADERBOARD = "leaderboard"
    CUSTOM = "custom"
    EXTERNAL = "external"


class CreateRegisteredModelVersionOperator(BaseDatarobotOperator):
    """
    Dynamically creates a registered model version using one of three methods:
    - Leaderboard Model
    - Custom Model Version
    - External Model

    :param model_version_params: Dictionary with parameters for creating model version.
    """

    template_fields = ["model_version_params"]

    def __init__(
        self,
        *,
        model_version_params: Dict[str, Any],
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.model_version_params = model_version_params
        self.model_creation_methods = {
            ModelType.LEADERBOARD: self.create_for_leaderboard,
            ModelType.CUSTOM: self.create_for_custom,
            ModelType.EXTERNAL: self.create_for_external,
        }

    def validate(self) -> None:
        """Validate required parameters before execution."""
        if not self.model_version_params.get("model_type"):
            raise ValueError("'model_type' must be specified in model_version_params.")

        try:
            model_type_enum = ModelType(self.model_version_params["model_type"])
        except ValueError:
            raise AirflowException(
                f"Invalid model_type: {self.model_version_params['model_type']}"
            ) from None

        if model_type_enum not in self.model_creation_methods:
            raise AirflowException(
                f"Unsupported model_type: {self.model_version_params['model_type']}"
            )

    def execute(self, context: Context) -> str:
        """Executes the operator to create a registered model version."""
        model_type_enum = ModelType(self.model_version_params["model_type"])
        create_method = self.model_creation_methods[model_type_enum]
        extra_params = {k: v for k, v in self.model_version_params.items() if k != "model_type"}
        version = create_method(**extra_params)
        self.log.info(f"Successfully created model version: {version.id}")
        return version.id

    def create_for_leaderboard(self, **kwargs):
        """Creates a registered model version from a leaderboard model."""
        return dr.RegisteredModelVersion.create_for_leaderboard_item(
            model_id=kwargs.pop("model_id"), **kwargs
        )

    def create_for_custom(self, **kwargs):
        """Creates a registered model version for a custom model."""
        return dr.RegisteredModelVersion.create_for_custom_model_version(
            custom_model_version_id=kwargs.pop("custom_model_version_id"), **kwargs
        )

    def create_for_external(self, **kwargs):
        """Creates a registered model version for an external model."""
        return dr.RegisteredModelVersion.create_for_external(
            target=kwargs.pop("target"), name=kwargs.pop("name"), **kwargs
        )



# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from datetime import datetime

from airflow.decorators import dag
from airflow.decorators import task

from datarobot_provider.operators.monitoring import GetMonitoringSettingsOperator
from datarobot_provider.operators.monitoring import UpdateMonitoringSettingsOperator


@dag(
    schedule=None,
    start_date=datetime(2023, 1, 1),
    tags=["example", "mlops"],
    # Default json config example:
    params={
        "target_drift_enabled": True,
        "feature_drift_enabled": True,
        "association_id_column": ["id"],
        "required_association_id": False,
        "predictions_data_collection_enabled": False,
    },
)
def deployment_monitoring_settings(deployment_id=None):
    if not deployment_id:
        raise ValueError("Invalid or missing `deployment_id` value")
    get_monitoring_settings_before_op = GetMonitoringSettingsOperator(
        task_id="get_monitoring_settings_before",
        # you can pass deployment_id from previous operator here:
        deployment_id=deployment_id,
    )

    update_monitoring_settings_op = UpdateMonitoringSettingsOperator(
        task_id="update_monitoring_settings",
        # you can pass deployment_id from previous operator here:
        deployment_id=deployment_id,
    )

    get_monitoring_settings_after_op = GetMonitoringSettingsOperator(
        task_id="get_monitoring_settings_after",
        # you can pass deployment_id from previous operator here:
        deployment_id=deployment_id,
    )

    @task(task_id="example_processing_python")
    def monitoring_settings_processing(
        model_monitoring_settings_before, model_monitoring_settings_after
    ):
        """Example of custom logic based on monitoring_settings from the deployment."""

        # Put your service stat processing logic here:

        settings_changed = {
            setting: model_monitoring_settings_after[setting]
            for setting in model_monitoring_settings_after
            if model_monitoring_settings_after[setting] != model_monitoring_settings_before[setting]
        }

        return settings_changed

    example_service_stat_processing = monitoring_settings_processing(
        model_monitoring_settings_before=get_monitoring_settings_before_op.output,
        model_monitoring_settings_after=get_monitoring_settings_after_op.output,
    )

    (
        get_monitoring_settings_before_op
        >> update_monitoring_settings_op
        >> get_monitoring_settings_after_op
        >> example_service_stat_processing
    )


deployment_monitoring_settings_dag = deployment_monitoring_settings()

if __name__ == "__main__":
    deployment_monitoring_settings_dag.test()



# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from typing import Any
from typing import List
from typing import Optional
from typing import Sequence

import datarobot as dr
from airflow.exceptions import AirflowException
from airflow.utils.context import Context
from datarobot import SegmentationTask
from datarobot.enums import DEFAULT_MAX_WAIT

from datarobot_provider.operators.base_datarobot_operator import BaseDatarobotOperator


class CreateSegmentationTaskOperator(BaseDatarobotOperator):
    """
    Creates a new segmentation task for a model in a project. This operator will handle the process
    of registering the model and constructing a segmentation task from the resulting registered model
    with the given parameters.The task can then be used to start a new autopilot using the segmentation task.

    See the StartAutopilotOperator and StartDatetimeAutopilotOperator for more information.

    Args:
        project_id (str): DataRobot project ID.
        target (str): Target variable for the segmentation task.
        use_time_series (bool): Whether to use time series data. Defaults to True.
        datetime_partition_column (Optional[str]): Datetime partition column name.
        multiseries_id_columns (Optional[List[str]]): List of multiseries ID column names.
        user_defined_segment_id_columns (Optional[List[str]]): List of user-defined segment ID column names.
        max_wait (int): Maximum wait time for the task to complete. Defaults to DEFAULT_MAX_WAIT.
        model_package_id (Optional[str]): Model package ID to use for the segmentation task.

    Returns:
        str: ID of the segmentation task created.
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = [
        "project_id",
        "model_id",
    ]

    def __init__(
        self,
        *,
        project_id: str,
        target: str,
        use_time_series: bool = True,
        datetime_partition_column: Optional[str] = None,
        multiseries_id_columns: Optional[List[str]] = None,
        user_defined_segment_id_columns: Optional[List[str]] = None,
        max_wait: int = DEFAULT_MAX_WAIT,
        model_package_id: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.target = target
        self.use_time_series = use_time_series
        self.datetime_partition_column = datetime_partition_column
        self.multiseries_id_columns = multiseries_id_columns
        self.user_defined_segment_id_columns = user_defined_segment_id_columns
        self.max_wait = max_wait
        self.model_package_id = model_package_id

    def validate(self) -> None:
        if not self.project_id:
            raise ValueError("project_id is required to create a SegmentationTask.")

        if not self.target:
            raise ValueError("target is required to create a SegmentationTask.")

    def execute(self, context: Context) -> str:
        segmentation_task_results = dr.SegmentationTask.create(
            project_id=self.project_id,
            target=self.target,
            use_time_series=self.use_time_series,
            datetime_partition_column=self.datetime_partition_column,
            multiseries_id_columns=self.multiseries_id_columns,
            user_defined_segment_id_columns=self.user_defined_segment_id_columns,
            max_wait=self.max_wait,
            model_package_id=self.model_package_id,
        )
        completed_jobs = segmentation_task_results["completedJobs"]
        if len(completed_jobs) == 0:
            raise AirflowException("No completed jobs found in segmentation task results.")

        segmentation_task: SegmentationTask = completed_jobs[0]
        return segmentation_task.id



# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.


DATAROBOT_CONN_ID = "datarobot_default"

XCOM_DEFAULT_USE_CASE_ID = "default_use_case_id"



# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
"""
Config example for this dag to run a Codespace notebook with notebook parameters:
{
    "notebook_id": "6798e158cc74377e3ceb4868",
    "notebook_path": "/home/notebooks/storage/test.py"
    "notebook_parameters": {
        "data": [{"name": "foo", "value": "bar"}]
    }
}
"""

from airflow.decorators import dag
from airflow.models.param import Param

from datarobot_provider._experimental.operators.notebook import NotebookRunOperator
from datarobot_provider._experimental.sensors.notebook import NotebookRunCompleteSensor

# Default to 24 hours
DEFAULT_NOTEBOOK_EXECUTION_TIMEOUT = 1440


@dag(
    schedule=None,
    description="Run a DataRobot notebook to completion and store a revision after it completes.",
    dag_display_name="Run DataRobot Notebook",
    tags=["example", "notebook"],
    params={
        "notebook_id": "put_your_notebook_id_here",
        "notebook_path": Param(None, type=["string", "null"]),
        "notebook_parameters": Param(None, type=["object", "null"]),
    },
    render_template_as_native_obj=True,
)
def datarobot_notebook_run():
    # Timeout waiting for complete notebook execution
    execution_timeout = DEFAULT_NOTEBOOK_EXECUTION_TIMEOUT * 60

    run_notebook = NotebookRunOperator(task_id="run_notebook")

    # Status check each 15 sec
    poke_interval = 15
    check_notebook_run_complete = NotebookRunCompleteSensor(
        task_id="check_notebook_run_complete",
        job_id=run_notebook.output,
        poke_interval=poke_interval,
        timeout=execution_timeout,
    )

    run_notebook >> check_notebook_run_complete


datarobot_notebook_run_dag = datarobot_notebook_run()

if __name__ == "__main__":
    datarobot_notebook_run_dag.test()



# Copyright 2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
"""
Config example for this dag:
{
    "deployment_id": "62cc0a6383d7a13d34f83344",
    "score_settings": {
        "intake_settings": {
            "type": "dataset",
            "dataset_id": "623d8ae79b186124a926c3cd"
        },
        "output_settings": {
            "type": "localFile",
            "path": "include/Diabetes_predictions.csv"
        }
    }
}
"""

from datetime import datetime

from airflow.decorators import dag

from datarobot_provider.operators.deployment import ScorePredictionsOperator
from datarobot_provider.sensors.datarobot import ScoringCompleteSensor


@dag(
    schedule=None,
    start_date=datetime(2022, 1, 1),
    tags=["example"],
)
def datarobot_score(deployment_id=None):
    if not deployment_id:
        raise ValueError("Invalid or missing `deployment_id` value")

    score_predictions_op = ScorePredictionsOperator(
        task_id="score_predictions",
        deployment_id=deployment_id,
    )

    scoring_complete_sensor = ScoringCompleteSensor(
        task_id="check_scoring_complete",
        job_id=score_predictions_op.output,
    )

    score_predictions_op >> scoring_complete_sensor


datarobot_pipeline_dag = datarobot_score()
